"""Default headers for session."""

SESSION_HEADERS = {
    'Content-type': 'application/json',
    'Accept': 'application/json',
    'User-Agent': 'ibs-jsonrpc',
    'Accept-Charset': 'utf-8',
    'Cache-Control': 'no-cache',
}



"""Get active leaves API method."""
from ibsng.handler.handler import Handler


class getActiveLeaves(Handler):
    """Get active leaves class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.order_by, str)
        self.is_valid(self.desc, bool)

    def setup(self, order_by, desc):
        """Setup required parameters.

        :param str order_by: order by a specific field
        :param bool desc: order of result

        :return: None
        :rtype: None
        """
        self.order_by = order_by
        self.desc = desc



"""IBSng 'method' exceptions handler."""
from .base import IBSngException


class ServerMethodError(IBSngException):
    """Raise when method result contains error."""

    _err_code = 401

    def __init__(self, error_text):
        """Raise when method result contains error.

        :param error_text: server error body
        :type error_text: str
        """
        message = "Check your entries, server method returned error: {}"
        message = message.format(error_text)
        super(ServerMethodError, self).__init__(message)


class ImportMethodNotFound(IBSngException):
    """Raise when method file not found."""

    _err_code = 402

    def __init__(self, method_name):
        """Raise when method file not found.

        :param method_name: method name
        :type method_name: str
        """
        message = "There is no '{}' method in driver.".format(method_name)
        super(ImportMethodNotFound, self).__init__(message)


class ImportMethodError(IBSngException):
    """Raise when importing method encountered with error."""

    _err_code = 403

    def __init__(self, method_name):
        """Raise when importing method encountered with error.

        :param method_name: method name
        :type method_name: str
        """
        message = "Import {} method in driver got error.".format(method_name)
        super(ImportMethodError, self).__init__(message)


class InvalidArgumentType(IBSngException):
    """Raise when method argument type is incorrect."""

    _err_code = 404

    def __init__(self, method_name, argument_value, argument_type):
        """Raise when method argument type is incorrect.

        :param str method_name: method name
        :param str argument_value: argument value
        :param str argument_type: argument type
        """
        message = "Method '{}' got incorrect argument type: '{}' ({})".\
                  format(method_name, argument_value, argument_type.__name__)
        super(InvalidArgumentType, self).__init__(message)


class InvalidArgumentValue(IBSngException):
    """Raise when method argument value is incorrect format."""

    _err_code = 405

    def __init__(self, method_name, argument_value, regex_pattern=""):
        """Raise when method argument value is incorrect format.

        :param str method_name: method name
        :param type argument_value: argument data
        :param type regex_pattern: regular expression pattern
        """
        if not regex_pattern:
            message = "Method '{}' got empty argument value: '{}'".\
                      format(method_name, argument_value)
        else:
            message = "Method '{}' got incorrect argument value: '{}' ({})".\
                      format(method_name, argument_value, regex_pattern)
        super(InvalidArgumentValue, self).__init__(message)



"""Kick idle users by ip API method."""
from ibsng.handler.handler import Handler


class kickIDLEUsersByIP(Handler):
    """Kick idle users by ip method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.idle_ip_list, list)

    def setup(self, idle_ip_list):
        """Setup required parameters.

        :param list idle_ip_list: list of ip addresses

        :return: None
        :rtype: None
        """
        self.idle_ip_list = idle_ip_list



"""Get invoice by profile id API method."""
from ibsng.handler.handler import Handler


class getInvoiceProfileByID(Handler):
    """Get invoice by profile id method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.profile_id, int)

    def setup(self, profile_id):
        """Setup required parameters.

        :param int profile_id: profile id

        :return: None
        :rtype: None
        """
        self.profile_id = profile_id



"""Search for OTPs API method."""
from ibsng.handler.handler import Handler


class searchOTPs(Handler):
    """Search for OTPs class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.conds, dict)
        self.is_valid(self.from_, int)
        self.is_valid(self.to_, int)
        self.is_valid(self.sort_by, str)
        self.is_valid(self.desc, bool)

    def setup(self, conds, from_, to_, sort_by, desc):
        """Setup required parameters.

        :param dict conds: conditions
        :param int _from: page start index
        :param int to: page end index
        :param choice sort_by: sort by a specific field
        :param bool desc: order of result

        :return: None
        :rtype: None
        """
        self.conds = conds
        self.from_ = from_
        self.to_ = to_
        self.sort_by = sort_by
        self.desc = desc



"""Add new admin API method."""
from ibsng.handler.handler import Handler


class addNewAdmin(Handler):
    """Add new admin class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.admin_username, str)
        self.is_valid(self.admin_password, str)
        self.is_valid(self.admin_isp_name, str)
        self.is_valid(self.name, str)
        self.is_valid(self.email, str)
        self.is_valid_content(self.email, self.EMAIL_PATTERN)
        self.is_valid(self.comment, str, False)
        self.is_valid(self.admin_has_otp, bool)
        self.is_valid(self.admin_request_limit, int)

    def setup(self, admin_username, admin_password, admin_isp_name, name,
              email, comment, admin_has_otp, admin_request_limit):
        """Setup required parameters.

        :param str admin_username: admin authentication username
        :param str admin_password: admin authentication password
        :param str admin_isp_name: admin isp name
        :param str name: admin real name
        :param str email: admin email
        :param str comment: comment about this action
        :param bool admin_has_otp: has admin otp
        :param int admin_request_limit: limitation on admin requests

        :return: None
        :rtype: None
        """
        self.admin_username = admin_username
        self.admin_password = admin_password
        self.admin_isp_name = admin_isp_name
        self.name = name
        self.email = email
        self.comment = comment
        self.admin_has_otp = admin_has_otp
        self.admin_request_limit = admin_request_limit



"""Search admin login history API method."""
from ibsng.handler.handler import Handler


class searchAdminLoginHistory(Handler):
    """Search admin login history method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.conds, dict)
        self.is_valid(self.from_index, int)
        self.is_valid(self.to_index, int)
        self.is_valid(self.sort_by, str)
        self.is_valid(self.desc, bool)

    def setup(self, conds, from_index, to_index, sort_by, desc):
        """Setup required parameters.

        :param dict conds: conditions
        :param int from_index: from index
        :param int to_index: end index
        :param str sort_by: sort result by specific field
        :param bool desc: descending

        :return: None
        :rtype: None
        """
        self.conds = conds
        self.__from = from_index
        self.to_ = to_index
        self.sort_by = sort_by
        self.desc = desc



"""Add new leaf API method."""
from ibsng.handler.handler import Handler


class addLeaf(Handler):
    """Add new leaf class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.leaf_name, str)
        self.is_valid(self.parent_id, int)
        self.is_valid_content(self.parent_id, self.ID_PATTERN)
        self.is_valid(self.default_rate_kbits, int)
        self.is_valid(self.default_ceil_kbits, int)
        self.is_valid(self.total_rate_kbits, int)
        self.is_valid(self.total_ceil_kbits, int)
        self.is_valid(self.default_priority, int)
        self.is_valid(self.total_priority, int)

    def setup(self, leaf_name, parent_id, default_rate_kbits,
              default_ceil_kbits, total_rate_kbits, total_ceil_kbits,
              default_priority, total_priority):
        """Setup required parameters.

        :param str leaf_name: leaf name
        :param int parent_id: parent leaf
        :param int default_rate_kbits: kilo bits
        :param int default_ceil_kbits: kilo bits
        :param int total_rate_kbits: kilo bits
        :param int total_ceil_kbits: kilo bits
        :param int default_priority: default priority
        :param int total_priority: total priority

        :return: None
        :rtype: None
        """
        self.leaf_name = leaf_name
        self.parent_id = parent_id
        self.default_rate_kbits = default_rate_kbits
        self.default_ceil_kbits = default_ceil_kbits
        self.total_rate_kbits = total_rate_kbits
        self.total_ceil_kbits = total_ceil_kbits
        self.default_priority = default_priority
        self.total_priority = total_priority



"""Change user status API method."""
from ibsng.handler.handler import Handler


class changeStatus(Handler):
    """Change user status method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.user_id, int)
        self.is_valid(self.status, str)

    def setup(self, user_id, status):
        """Setup required parameters.

        :param str user_id: ibsng user id
        :param str status: new user status (Package, Recharged,
                           Temporary Extended)

        :return: None
        :rtype: None
        """
        self.user_id = user_id
        self.status = status



"""Delete bandwidth static IP API method."""
from ibsng.handler.handler import Handler


class delBwStaticIP(Handler):
    """Delete bandwidth static IP class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.ip_addr, str)
        self.is_valid_content(self.ip_addr, self.IP_PATTERN)

    def setup(self, ip_addr):
        """Setup required parameters.

        :param str ip_addr: ip address

        :return: None
        :rtype: None
        """
        self.ip_addr = ip_addr



"""String library.

This library provides some useful functions which works on strings.
"""
from ibsng.exception import connection
import re


def connection_string_parser(conn_str):
    """Parse connection string pattern.

    :raises ValueError: raise ValueError when can't parse connection string
    :raises InvalidConnectionString: rasie InvalidConnectionString when
                                     connection string is not in correct style

    :param conn_str: connection string
    :type conn_str: str

    :return: connection config
    :rtype: tuple
    """
    try:
        # Fetch server info (host/port)
        if conn_str.rfind('@') >= 0:
            server_url = conn_str[conn_str.rindex('@')+1:]
        else:
            raise ValueError("Invalid connection string structure.")

        if not validate_server_addr(server_url):
            raise ValueError("Invalid server address.")

        # Fetch IP address
        host = server_url.split(':')[1][2:]

        # Fetch user info (username/password)
        user_info = conn_str[:conn_str.rfind('@')]

        # Prepare result
        auth_params = {
            'auth_remoteaddr': host,
            'auth_name': user_info[:user_info.find(':')],
            'auth_pass': user_info[user_info.find(':')+1:],
            'login_auth_name': user_info[:user_info.find(':')],
            'login_auth_pass': user_info[user_info.find(':')+1:],
        }
        return (auth_params, server_url)
    except Exception:
        raise connection.InvalidConnectionString()


def camel_to_snake(string):
    """Convert camel to snake style.

    :param string: camel string
    :type string: str

    :return: snake style string
    :rtype: str
    """
    string = "".join([(x.replace(x, '_' + x.lower() if x.isupper() else x))
                     for x in list(string)])
    if string.startswith('_'):
        string = string[1:]
    return string


def validate_server_addr(server_addr):
    """Server address validation.

    :param server_addr: Server address (domain name or IP)
    :type server_addr: str

    :return: return validation result
    :rtype: bool
    """
    check_pattern = "^(http|https)://([0-9]+\.[0-9]+\.[0-9]+\.[0-9]+|" + \
                    "[a-zA-Z0-9\-\.]+\.[a-zA-Z]+)(:[0-9]+)?$"
    if not re.match(check_pattern, server_addr):
        return False
    return True



"""Get all permissions API method."""
from ibsng.handler.handler import Handler


class getAllPerms(Handler):
    """Get all permissions method class."""

    def control(self):
        """Validate inputs after method setup.

        :return: None
        :rtype: None
        """
        if hasattr(self, 'category'):
            self.is_valid(self.category, str)

    def setup(self, category):
        """Setup required parameters.

        :param type category: category

        :return: None
        :rtype: None
        """
        if category:
            self.category = category



"""Get user info API method."""
from ibsng.handler.handler import Handler


class getUserInfo(Handler):
    """Get user info method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.username, str)

    def setup(self, username):
        """Setup required parameters.

        :param str username: ibsng username

        :return: None
        :rtype: None
        """
        self.username = username



"""Add new user API method."""
from ibsng.handler.handler import Handler


class addNewUsers(Handler):
    """Add new user method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.count, int)
        self.is_valid(self.credit, float)
        self.is_valid(self.isp_name, str)
        self.is_valid(self.group_name, str)
        self.is_valid(self.credit_comment, str)
        self.is_valid(self.custom_fields, dict, False)

    def setup(self, count, credit, isp_name, group_name,
              credit_comment="", custom_fields={}):
        """Setup required parameters.

        :param int count: count of users
        :param float credit: initialize credit
        :param str isp_name: isp name
        :param str group_name: group name
        :param str credit_comment: credit comment
        :param dict custom_fields: custom fields

        :return: None
        :rtype: None
        """
        self.count = count
        self.credit = credit
        self.isp_name = isp_name
        self.group_name = group_name
        self.credit_comment = credit_comment
        self.custom_fields = custom_fields



"""Get session isp id API method."""
from ibsng.handler.handler import Handler


class getSessionISPID(Handler):
    """Get session isp id method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.web_domain, str)

    def setup(self, web_domain):
        """Setup required parameters.

        :param str web_domain: domain of user/admin panel page

        :return: None
        :rtype: None
        """
        self.web_domain = web_domain



"""Update charge API method."""
from ibsng.handler.handler import Handler


class updateCharge(Handler):
    """Update charge method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.charge_id, int)
        self.is_valid(self.charge_name, str)
        self.is_valid(self.isp_name, str)
        self.is_valid(self.comment, str, False)

    def setup(self, charge_id, charge_name, isp_name, comment=""):
        """Setup required parameters.

        :param int charge_id: charge id
        :param str charge_name: new charge name
        :param str isp_name: new isp name
        :param str comment: new charge comment

        :return: None
        :rtype: None
        """
        self.charge_id = charge_id
        self.charge_name = charge_name
        self.comment = comment
        self.isp_name = isp_name



"""AFE change user credit API method."""
from ibsng.handler.handler import Handler


class afeChangeUserCredit(Handler):
    """AFE change user credit method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.username, str)
        self.is_valid(self.amount, float)
        self.is_valid(self.comment, str, False)

    def setup(self, username, amount, comment=""):
        """Setup required parameters.

        :param str username: ibsng username
        :param float amount: user amount
        :param str comment: comment for this action

        :return: None
        :rtype: None
        """
        self.username = username
        self.amount = amount
        self.comment = comment



"""Get voucher by id API method."""
from ibsng.handler.handler import Handler


class getVoucherByID(Handler):
    """Get voucher by id method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.voucher_id, int)

    def setup(self, voucher_id):
        """Setup required parameters.

        :param int voucher_id: voucher id

        :return: None
        :rtype: None
        """
        self.voucher_id = voucher_id



"""Bulk action to delete user API method."""
from ibsng.handler.handler import Handler


class bulkDeleteUser(Handler):
    """Bulk action to delte user method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.conds, dict)
        self.is_valid(self.del_connection_logs, bool)
        self.is_valid(self.del_audit_logs, bool)
        self.is_valid(self.delete_comment, str, False)

    def setup(self, conds, del_connection_logs,
              del_audit_logs, delete_comment=""):
        """Setup required parameters.

        :param dict conds: conditions
        :param bool del_connection_logs: delete user connection logs
        :param bool del_audit_logs: delete user audit logs
        :param str delete_comment: set comment for this action

        :return: None
        :rtype: None
        """
        self.conds = conds
        self.del_connection_logs = del_connection_logs
        self.del_audit_logs = del_audit_logs
        self.delete_comment = delete_comment



"""Change credit extended API method."""
from ibsng.handler.handler import Handler


class changeCreditExtended(Handler):
    """Change credit extended method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.user_id, int)
        self.is_valid(self.credit, float)
        self.is_valid(self.change_type, str)
        self.is_valid_content(self.change_type, ("ADD|SET|MULTIPLY"))
        self.is_valid(self.credit_comment, str, False)

    def setup(self, user_id, credit, change_type, credit_comment=""):
        """Setup required parameters.

        :param str user_id: ibsng user id
        :param float credit: new credit value
        :param str change_type: type of change (ADD, SET, MULTIPLY)
        :param str credit_comment: comment for this credit

        :return: None
        :rtype: None
        """
        self.user_id = user_id
        self.credit = credit
        self.change_type = change_type
        self.credit_comment = credit_comment



"""Main session handler library.

This library prepares session id for requests.
"""
import requests
from ibsng.setting import session as session_settings


class Session:
    """Session handler class."""

    _is_instance = None

    def __init__(self):
        """Session handler init."""
        self._session = requests.sessions.Session()
        self._session.headers.update(session_settings.SESSION_HEADERS)

    def __new__(cls, *args, **kwargs):
        """Singletone method to generate session."""
        if not cls._is_instance:
            cls._is_instance = object.__new__(Session, *args, **kwargs)
        return cls._is_instance

    def get_session(self):
        """Return current session.

        :return: current session
        :rtype: str
        """
        return self._session



"""Hash library.

This library provides some functions which are useful
to work on dict/json objects.
"""
import json


def serialize(data):
    """Serialize dict object.

    :raises ValueError: raise ValueError when can't dump dict

    :param data: dict object
    :type data: dict

    :return: json string
    :rtype: str
    """
    return json.dumps(data)


def deserialize(data):
    """Deserialize json string.

    :raises ValueError: raise ValueError when can't load json

    :param data: json string
    :type data: str

    :return: dict object
    :rtype: dict
    """
    try:
        return json.loads(data)
    except (json.JSONDecodeError, TypeError):
        raise ValueError("Deserialize failed. Invalid data.")


class dict_to_object:
    """Convert dictionary to object."""

    def __init__(self, data_dict):
        """Convert dictionary to object.

        :param data_dict: data dict object
        :type data_dict: dict
        """
        for k, v in data_dict.items():
            if (isinstance(v, (list, tuple))):
                setattr(self, k, [dict_to_object(x)
                                  if isinstance(x, dict) else x for x in v])
            else:
                setattr(self, k, dict_to_object(v)
                        if isinstance(v, dict) else v)

    def inspect(self, key_name=None):
        """Prepare result as readable structure.

        :param key_name: result key name
        :type key_name: str

        :return: readable data structure
        :rtype: object
        """
        result = []
        for var in dir(self):
            if var.startswith("__"):
                continue

            if key_name and key_name == var:
                return getattr(self, var)

            result.append(var)
        return result



"""Delete ras API method."""
from ibsng.handler.handler import Handler


class deleteRas(Handler):
    """Delete ras method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.ras_id, int)

    def setup(self, ras_id):
        """Setup required parameters.

        :param int ras_id: ras id

        :return: None
        :rtype: None
        """
        self.ras_id = ras_id



"""Get batch voucher attributes API method."""
from ibsng.handler.handler import Handler


class getBatchVoucherAttrs(Handler):
    """Get batch voucher attributes method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.voucher_pin, str)

    def setup(self, voucher_pin):
        """Setup required parameters.

        :param str voucher_pin: voucher pin

        :return: None
        :rtype: None
        """
        self.voucher_pin = voucher_pin



"""Add new charge rule API method."""
from ibsng.handler.handler import Handler


class addNewChargeRule(Handler):
    """Add new charge rule method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.charge_rule_description, str, False)
        self.is_valid(self.charge_rule_priority, int)
        self.is_valid(self.charge_name, str)

    def setup(self, charge_rule_description,
              charge_rule_priority, charge_name):
        """Setup required parameters.

        :param str charge_rule_description: charge rule description (comment)
        :param int charge_rule_priority: charge rule priority
        :param str charge_name: charge name

        :return: None
        :rtype: None
        """
        self.charge_rule_description = charge_rule_description
        self.charge_rule_priority = charge_rule_priority
        self.charge_name = charge_name



"""Get ISP info API method."""
from ibsng.handler.handler import Handler


class getISPInfo(Handler):
    """Get ISP info method class."""

    def control(self):
        """Validate inputs after method setup.

        :return: None
        :rtype: None
        """
        self.is_valid(self.isp_name, str, False)
        self.is_valid(self.isp_id, int, False)

    def setup(self, isp_name=None, isp_id=None):
        """Setup required parameters.

        :param str isp_name: isp name
        :param int isp_id: isp id

        :return: None
        :rtype: None
        """
        self.isp_name = isp_name
        self.isp_id = isp_id



"""Temporary extend users API method."""
from ibsng.handler.handler import Handler


class temporaryExtendUsers(Handler):
    """Temporary extend users method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.user_id, str)
        self.is_valid_content(self.user_id, self.IDS_PATTERN)
        self.is_valid(self.temporary_extend_hours, int)
        self.is_valid(self.temporary_extend_credit, float)

    def setup(self, user_ids, temporary_extend_hours, temporary_extend_credit):
        """Setup required parameters.

        :param list user_ids: ibsng user ids
        :param int temporary_extend_hours: hours
        :param float temporary_extend_credit: credit

        :return: None
        :rtype: None
        """
        self.user_id = ",".join(map(str, user_ids))
        self.temporary_extend_hours = temporary_extend_hours
        self.temporary_extend_credit = temporary_extend_credit



"""Update group attributes API method."""
from ibsng.handler.handler import Handler


class updateGroupAttrs(Handler):
    """Update group attributes method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.group_name, int)
        self.is_valid(self.attrs, dict, False)
        self.is_valid(self.to_del_attrs, list, False)

    def setup(self, group_name, attrs, to_del_attrs):
        """Setup required parameters.

        :param str group_name: group name
        :param dict attrs: updated attributes
        :param list to_del_attrs: deleted attributes

        :return: None
        :rtype: None
        """
        self.group_name = group_name
        self.attrs = attrs
        self.to_del_attrs = to_del_attrs



"""Get leaf info API method."""
from ibsng.handler.handler import Handler


class getLeafInfo(Handler):
    """Get leaf info class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.leaf_name, str)

    def setup(self, leaf_name):
        """Setup required parameters.

        :param str leaf_name: leaf name

        :return: None
        :rtype: None
        """
        self.leaf_name = leaf_name



"""Connection handler library.

This library is middleware between request and session libraries.
"""
from .request import Request
from .session import Session
from ibsng.util import string, file_dir
from ibsng.setting import auth as auth_setting
from ibsng.setting import default as default_setting
from ibsng.exception import (response as response_exception,
                             handler as handler_exception,
                             method as method_exception,
                             general as general_exception)
from ibsng import handler
from ibsng.util import hash


class Connection:
    """Connection handler class."""

    def __init__(self, conn_str, timeout=0.5, **headers):
        """Parse connection string and prepare session.

        :raises ValueError: raise ValueError when connection string is empty or
                            server address is not valid

        :param conn_str: connection string
        :type conn_str: str
        :param **headers: headers which should be overrided with main headers
        :type **headers: dict
        """
        if not conn_str:
            raise ValueError("Invalid connection string structure.")

        # required variables for requests
        self._headers = None
        self._auth_params, self._server_url = \
            string.connection_string_parser(conn_str)

        self._session = Session().get_session()
        self._request = Request(self._server_url,
                                self._session,
                                timeout=timeout)

        self.handler = handler

        # ready state to send requests
        self._r_steps = 0
        self._r_handler_name = None
        self._r_method_name = None
        self._r_method_module_name = None

        # list of all items in chain just for debugging
        self._r_chain = []

        # update headers
        self._update_headers(**headers)

    def _reset_variables(self, keep_chain=False):
        """Reset required variables.

        :param keep_chain: keep chain for debugging
        :type keep_chain: bool

        :return: None
        :rtype: None
        """
        self._r_steps = 0
        self._r_handler_name = None
        self._r_method_name = None
        if not keep_chain:
            self._r_chain.clear()

    def _update_headers(self, **headers):
        """Update headers.

        :param **headers: all required headers which should be overrided
        :type headers: dict

        :return: None
        :rtype: None
        """
        for k, v in headers.items():
            if default_setting.DEFAULT_HEADERS.get(k):
                default_setting.DEFAULT_HEADERS[k] = v
            else:
                default_setting.DEFAULT_HEADERS.update({k: v})
        self._headers = default_setting.DEFAULT_HEADERS
        # Add auth headers
        for k, v in auth_setting.AUTH_HEADERS.items():
            if not self._headers.get(k):
                self._headers.update({k: v})

    def auth(self):
        """Authenticate to the server.

        It tries to authenticate user and updates headers.

        :return: None
        :rtype: None
        """
        # Prepare required headers for authentication
        tmp_params = self._auth_params
        for k, v in self._headers.items():
            if not tmp_params.get(k):
                tmp_params.update({k: v})

        # Call login.login method with parameters
        action = self._call("login.login", **tmp_params)
        self._headers.update({"auth_session": action.get("result")})
        self._headers.update({"auth_name": self._auth_params["auth_name"]})
        self.clean_headers()
        self._r_chain = []
        self._r_handler_name = None

    def clean_headers(self):
        """Remove some headers after successful authentication.

        :return: None
        :rtype: None
        """
        self._headers.pop('create_session', None)
        self._headers.pop('login_auth_type', None)
        self._headers.pop('login_auth_pass', None)
        self._headers.pop('auth_pass', None)

    def _call(self, method, **params):
        """Send request to the server.

        :param method: API handler and method names
        :type method: str
        :param **params: method parameters
        :type **params: dict

        :return: server response
        :rtype: object
        """
        action = self._request.send(method, **params)
        if action.get("error"):
            raise response_exception.throw_exception(action["error"])
        return action

    def __call__(self, *args, **kwargs):
        """Call handler method."""
        # Check handler existance
        if not file_dir.is_handler_exist(self._r_handler_name):
            raise handler_exception.ImportHandlerNotFound(self._r_handler_name)

        # Check method existance
        if not file_dir.is_method_exist(self._r_handler_name,
                                        self._r_method_module_name):
            raise method_exception.ImportMethodNotFound(self._r_method_name)

        # Import handler/method module
        try:
            module = "ibsng.handler.{}.{}". \
                     format(self._r_handler_name, self._r_method_module_name)
            imp = __import__(module, fromlist=self._r_method_name)
            action = getattr(imp, self._r_method_name)(*args, **kwargs)
        except ImportError:
            raise method_exception.ImportMethodError(self._r_method_name)

        # Prepare headers
        tmp_headers = self._headers.copy()
        tmp_headers.update(action.serialize())

        # It's time to call server method
        result = self._call('{}.{}'.format(self._r_handler_name,
                                           self._r_method_name),
                            **tmp_headers)

        self._reset_variables()
        action.outcome(result)
        json_result = result
        result = hash.dict_to_object(result)
        setattr(result, "json", json_result)
        return result

    def __getattr__(self, api_name):
        """Handler/Method handler.

        Handle first(handler) and second(method) calls.

        .. raises:: ImportMethodNotFound: raise ImportMethodNotFound when
            method file not found
        .. raises:: ImportHandlerNotFound: raise ImportHandlerNotFound when
            handler directory not found
        .. raises:: ImportMethodError: raise ImportMethodError when importing
            method encoured with erorr
        .. raises:: general_exception.ChainFormatError: in case on any invalid
            chain structure

        :param api_name: handler or method name
        :type api_name: str
        """
        self._r_chain.append(api_name)
        if self._r_steps == 0:
            self._r_handler_name = api_name
        elif self._r_steps == 1:
            self._r_method_name = api_name
            self._r_method_module_name = string.camel_to_snake(api_name)
        else:
            self._reset_variables(True)
            raise general_exception.ChainFormatError(self._r_chain)

        self._r_steps += 1
        return self



"""Add new ISP API method."""
from ibsng.handler.handler import Handler


class addNewISP(Handler):
    """Add new ISP method class."""

    def control(self):
        """Validate inputs after method setup.

        :return: None
        :rtype: None
        """
        self.is_valid(self.isp_name, str)
        self.is_valid(self.parent_isp_name, str)
        self.is_valid(self.isp_has_deposit_limit, bool)
        self.is_valid(self.isp_deposit, float)
        self.is_valid(self.isp_mapped_user_id, int)
        self.is_valid(self.isp_auth_domain, str)
        self.is_valid(self.isp_web_domain, str)
        self.is_valid(self.isp_email, str)
        self.is_valid(self.prevent_neg_deposit_login, bool)
        self.is_valid(self.isp_comment, str)
        self.is_valid(self.isp_failed_user_id, int)

    def setup(self, isp_name, parent_isp_name, isp_has_deposit_limit,
              isp_deposit, isp_mapped_user_id, isp_auth_domain, isp_web_domain,
              isp_email, prevent_neg_deposit_login,
              isp_comment, isp_failed_user_id):
        """Setup required parameters.

        :param str isp_name: isp name
        :param str parent_isp_name: parent isp name
        :param bool isp_has_deposit_limit: isp has deposit limit
        :param float isp_deposit: isp deposit
        :param int isp_mapped_user_id: isp mapped user id
        :param str isp_auth_domain: isp auth domain
        :param str isp_web_domain: isp web domain
        :param str isp_email: isp email
        :param bool prevent_neg_deposit_login: prevent neg deposit login
        :param str isp_comment: isp comment
        :param int isp_failed_user_id: isp failed user id

        :return: None
        :rtype: None
        """
        self.isp_name = isp_name
        self.parent_isp_name = parent_isp_name
        self.isp_has_deposit_limit = isp_has_deposit_limit
        self.isp_deposit = isp_deposit
        self.isp_mapped_user_id = isp_mapped_user_id
        self.isp_auth_domain = isp_auth_domain
        self.isp_web_domain = isp_web_domain
        self.isp_email = isp_email
        self.prevent_neg_deposit_login = prevent_neg_deposit_login
        self.isp_comment = isp_comment
        self.isp_failed_user_id = isp_failed_user_id



"""Web login API method."""
from ibsng.handler.handler import Handler


class webLogin(Handler):
    """Web login method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.login_auth_type, str)
        self.is_valid_choice(self.login_auth_type,
                             ["ADMIN", "NORMAL_USER", "VOIP_USER"])
        self.is_valid(self.login_auth_name, str)
        self.is_valid(self.login_auth_pass, str)
        self.is_valid(self.auth_remoteaddr, str)
        self.is_valid_content(self.auth_remoteaddr, self.IP_PATTERN)

    def setup(self, login_auth_type, login_auth_name,
              login_auth_pass, auth_remoteaddr):
        """Setup required parameters.

        :param str login_auth_type: authentication type
                                    ("ADMIN", "NORMAL_USER", "VOIP_USER")
        :param str login_auth_name: authentication username
        :param str login_auth_pass: authentication password
        :param str auth_remoteaddr: remote IP of requester(admin/user)

        :return: None
        :rtype: None
        """
        self.login_auth_type = login_auth_type
        self.login_auth_name = login_auth_name
        self.login_auth_pass = login_auth_pass
        self.auth_remoteaddr = auth_remoteaddr



"""Get all ISP infos API method."""
from ibsng.handler.handler import Handler


class getAllISPInfos(Handler):
    """Get all ISP infos method class."""

    def control(self):
        """Validate inputs after method setup.

        :return: None
        :rtype: None
        """
        if hasattr(self, 'sort_by'):
            self.is_valid(self.sort_by, str)

    def setup(self, sort_by=None):
        """Setup required parameters.

        :param type sort_by: sort by

        :return: None
        :rtype: None
        """
        if sort_by:
            self.sort_by = sort_by



"""Get ISP tree API method."""
from ibsng.handler.handler import Handler


class getISPTree(Handler):
    """Get ISP tree method class."""

    def control(self):
        """Validate inputs after method setup.

        :return: None
        :rtype: None
        """
        self.is_valid(self.isp_name, str, False)

    def setup(self, isp_name=None):
        """Setup required parameters.

        :param str isp_name: isp name (if this param is missing, we use
                             admin's owner ISP)

        :return: None
        :rtype: None
        """
        self.isp_name = isp_name



"""Voucher get batch by id API method."""
from ibsng.handler.handler import Handler


class voucherGetBatchByID(Handler):
    """Voucher get batch by id method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.batch_id, str)

    def setup(self, batch_id):
        """Setup required parameters.

        :param int batch_id: batch id

        :return: None
        :rtype: None
        """
        self.batch_id = batch_id



"""Get active ras by ips API method."""
from ibsng.handler.handler import Handler


class getActiveRasIPs(Handler):
    """Get active ras by ips info method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.type_filter, str)

    def setup(self, type_filter):
        """Setup required parameters.

        :param str type_filter: type filter

        :return: None
        :rtype: None
        """
        self.type_filter = type_filter



"""IBSng 'connection' exceptions handler."""
from .base import IBSngException


class InvalidConnectionString(IBSngException):
    """Raise when connection string pattern is not correct."""

    _err_code = 100

    def __init__(self):
        """Raise when connection string pattern is not correct."""
        message = "Invalid connection string structure. \
                   Structure is <host:port>@<username:password>"
        super(InvalidConnectionString, self).__init__(message)


class ServerConnectionError(IBSngException):
    """Raise when can't connect to the server."""

    _err_code = 101

    def __init__(self, msg=None):
        """Raise when can't connect to the server.

        :param msg: message body
        :type msg: str
        """
        if not msg:
            message = "Server connection failed. Check your connection."
        else:
            message = msg
        super(ServerConnectionError, self).__init__(message)


class ServerConnectionTimeout(IBSngException):
    """Raise when can't connect to the server."""

    _err_code = 102

    def __init__(self):
        """Raise when connecting to the server gets timeout."""
        message = "Server connection timeout. Check your connection."
        super(ServerConnectionTimeout, self).__init__(message)



"""Delete charge API method."""
from ibsng.handler.handler import Handler


class deleteCharge(Handler):
    """Delete charge method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.charge_name, str)

    def setup(self, charge_name):
        """Setup required parameters.

        :param str charge_name: charge name

        :return: None
        :rtype: None
        """
        self.charge_name = charge_name



"""Get charge apps by charge id API method."""
from ibsng.handler.handler import Handler


class getAppsByChargeId(Handler):
    """Get charge apps by charge id method."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.charge_id, int)

    def setup(self, charge_id):
        """Setup required parameters.

        :param int charge_id: charge id

        :return: None
        :rtype: None
        """
        self.charge_id = charge_id



"""Update extra charge API method."""
from ibsng.handler.handler import Handler


class updateExtraChargeProfile(Handler):
    """Update extra charge method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.profile_id, int)
        self.is_valid(self.profile_name, str)
        self.is_valid(self.effective_hour, str)
        self.is_valid(self.rules, list)
        self.is_valid(self.comment, str, False)

    def setup(self, profile_id, profile_name,
              effective_hour, rules, comment=""):
        """Setup required parameters.

        :param int profile_id: profile id
        :param str profile_name: new profile name
        :param str effective_hour: new effective hour (format: HH:MM)
        :param list rules: new rules list
        :param str comment: new extra charge comment

        :return: None
        :rtype: None
        """
        self.profile_id = profile_id
        self.profile_name = profile_name
        self.effective_hour = effective_hour
        self.rules = rules
        self.comment = comment



"""Reset ISP page style API method."""
from ibsng.handler.handler import Handler


class resetISPPageStyle(Handler):
    """Reset ISP page style method class."""

    def control(self):
        """Validate inputs after method setup.

        :return: None
        :rtype: None
        """
        self.is_valid(self.isp_id, int)

    def setup(self, isp_id):
        """Setup required parameters.

        :param int isp_id: isp id

        :return: None
        :rtype: None
        """
        self.isp_id = isp_id



"""Main inputs validator library."""
import re
from abc import ABCMeta, abstractmethod

from ibsng.exception.method import InvalidArgumentType, InvalidArgumentValue


class Control(object, metaclass=ABCMeta):
    """Controlling inputs.

    .. note: do not implement __init__ method for this class.
    """

    # useful patterns
    EMAIL_PATTERN = r"[a-zA-Z0-9\_\-\.\+]+@[a-zA-Z0-9\.\-]+\.[a-zA-Z]+"
    IP_PATTERN = r"^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$"
    ID_PATTERN = r"^[0-9]+$"
    IDS_PATTERN = r"^[0-9\,]+$"
    POSITIVE_NUMBER = r"^[0-9]+$"

    def is_valid(self, argument_value, argument_type, value_check=True):
        """Validate argument type before sending to the server.

        .. raise: InvalidArgumentType

        :param object argument_value: argument data
        :param type argument_type: type of argument which argument must be.

        :return: argument validation
        :rtype: bool
        """
        if value_check and not argument_value:
            raise InvalidArgumentValue(self.__class__.__name__,
                                       argument_value)

        if not isinstance(argument_value, argument_type):
            raise InvalidArgumentType(self.__class__.__name__,
                                      argument_value,
                                      argument_type)
        return True

    def is_valid_content(self, argument_value, regex_pattern):
        """Validate argument content before sending to the server.

        .. raise: InvalidArgumentValue

        :param object argument_value: argument data
        :param str regex_pattern: regular expression pattern

        :return: argument validation
        :rtype: bool
        """
        is_valid = True

        if isinstance(argument_value, (list, tuple)):
            for value in argument_value:
                if not value or not re.match(regex_pattern, value):
                    is_valid = False
                    break
        else:
            if not argument_value or \
               not re.match(regex_pattern, argument_value):
                is_valid = False

        if not is_valid:
            raise InvalidArgumentValue(self.__class__.__name__,
                                       argument_value,
                                       regex_pattern)
        return True

    def is_valid_choice(self, argument_value, valid_list):
        """Validate argument content based on valid list.

        .. raise: InvalidArgumentValue

        :param object argument_value: argument data
        :param list valid_list: valid data list

        :return: argument validation
        :rtype: bool
        """
        if argument_value not in valid_list:
            raise InvalidArgumentValue(self.__class__.__name__,
                                       argument_value,
                                       valid_list)
        return True

    @abstractmethod
    def control(self):
        """Validate arguments of methods (input).

        .. note: Should be implemented based on requirements

        :return: None
        :rtype: None
        """
        pass



"""Initialize handler methods.

This class will not invoke by someone. It's just for inheritance.
"""
from abc import ABCMeta, abstractmethod


class Initiator(object, metaclass=ABCMeta):
    """Initialize handler method."""

    def __init__(self, *args, **kwargs):
        """Make headers ready."""
        self.setup(*args, **kwargs)

    @abstractmethod
    def setup(self, *args, **kwargs):
        """Set up passed arguments and update method headers."""
        raise NotImplementedError()



"""Get group info API method."""
from ibsng.handler.handler import Handler


class getGroupInfo(Handler):
    """Get group info method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.group_name, str, False)
        self.is_valid(self.group_id, int, False)

    def setup(self, group_name=None, group_id=None):
        """Setup required parameters.

        :param str group_name: group name
        :param str group_id: group id

        :return: None
        :rtype: None
        """
        if group_name:
            self.group_name = group_name
        elif group_id:
            self.group_id = group_id



"""Run debug code API method."""
from ibsng.handler.handler import Handler


class runDebugCode(Handler):
    """Run debug code method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.command, str)

    def setup(self, command):
        """Setup required parameters.

        :param str command: command

        :return: None
        :rtype: None
        """
        self.command = command



"""Recharge users API method."""
from ibsng.handler.handler import Handler


class rechargeUsers(Handler):
    """Recharge users method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.user_id, str)
        self.is_valid_content(self.user_id, self.IDS_PATTERN)
        self.is_valid(self.comment, str, False)

    def setup(self, user_ids, comment=""):
        """Setup required parameters.

        :param list user_ids: ibsng user ids
        :param str comment: comment for this action

        :return: None
        :rtype: None
        """
        self.user_id = ",".join(map(str, user_ids))
        self.comment = comment



"""Util for file and directory.

This library provides some functions to work on file and directory.
"""
import os
from ibsng.setting import general
from ibsng.util import string


def is_handler_exist(handler_name):
    """Check the handler is exist in driver.

    :param handler_name: handler name
    :type handler_name: str

    :return: is handler exists or not
    :rtype: bool
    """
    handler_path = os.path.join(general.BASE_DIR,
                                "handler/{}".format(handler_name))
    if os.path.exists(handler_path):
        return True
    return False


def is_method_exist(handler_name, method_name):
    """Check the method is exist in driver.

    :param handler_name: handler name
    :type handler_name: str
    :param metho_name: method name
    :type method_name: str

    :return: is method exists or not
    :rtype: bool
    """
    method_name = string.camel_to_snake(method_name)
    method_path = os.path.join(general.BASE_DIR,
                               "handler/{}/{}.py".format(handler_name,
                                                         method_name))
    if os.path.exists(method_path):
        return True
    return False



"""List charges API method."""
from ibsng.handler.handler import Handler


class listCharges(Handler):
    """List charges method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.isp_id, int)

    def setup(self, isp_id):
        """Setup required parameters.

        :param str isp_id: comma-seprated. list of isp ids

        :return: None
        :rtype: None
        """
        self.isp_id = isp_id



"""Update ras info API method."""
from ibsng.handler.handler import Handler


class updateRasInfo(Handler):
    """Update ras info method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.ras_id, int)
        self.is_valid(self.ras_ip, str)
        self.is_valid_content(self.ras_ip, self.IP_PATTERN)
        self.is_valid(self.ras_type, str)
        self.is_valid(self.radius_secret, str)
        self.is_valid(self.ras_description, str)
        self.is_valid(self.comment, str, False)

    def setup(self, ras_id, ras_ip, ras_type,
              radius_secret, ras_description, comment=""):
        """Setup required parameters.

        :param int ras_id: ras id
        :param str ras_ip: ras ip address
        :param str ras_type: ras type (get from ras.getRasTypes method)
        :param str radius_secret: radius secret key
        :param str ras_description: ras description
        :param str comment: comment

        :return: None
        :rtype: None
        """
        self.ras_id = ras_id
        self.ras_ip = ras_ip
        self.ras_type = ras_type
        self.radius_secret = radius_secret
        self.ras_description = ras_description
        self.comment = comment



"""Get users with expiration date and first login API method."""
from ibsng.handler.handler import Handler


class getUsersExpDateFirstLogin(Handler):
    """Get users with expiration date and first login method class."""

    pass



"""IBSng 'response exception' exceptions handler."""
from . import auth, unknown


def throw_exception(server_error):
    """Throw exception based on server error.

    :param server_error: sever error body
    :type server_error: str

    :return: exception class
    :rtype: object
    """
    error_code = 'UNKNOWN'

    # Check error content contains error code
    if server_error.find('|') >= 0:
        error_code = server_error.split('|')[0]

    # Find matched exception
    e = {
        'INVALID_ADMIN_NAME': auth.InvalidAuthentication,
        'INCORRECT_PASSWORD': auth.InvalidAuthentication,
        'INVALID_SESSION_ID': auth.InvalidSessionId,
        'UNKNOWN': unknown.UnknownError,
    }.get(error_code, unknown.UnknownError)

    if error_code == 'UNKNOWN' or e == unknown.UnknownError:
        return e(server_error)
    return e()



"""Add new leaf service API method."""
from ibsng.handler.handler import Handler


class addLeafService(Handler):
    """Add new leaf service class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.leaf_name, str)
        self.is_valid(self.dst_ip, str)
        self.is_valid_content(self.dst_ip, self.IP_PATTERN)
        self.is_valid(self.leaf_name, str)
        self.is_valid(self.protocol, str)
        self.is_valid_content(self.protocol, "(tcp|udp|icmp|ip)")
        self.is_valid(self.filter_, str)
        self.is_valid(self.rate_kbits, int)
        self.is_valid(self.ceil_kbits, int)
        self.is_valid(self.priority, int)
        self.is_valid_content(self.priority, self.POSITIVE_NUMBER)

    def setup(self, leaf_name, dst_ip, protocol, filter_format,
              rate_kbits, ceil_kbits, priority):
        """Setup required parameters.

        :param str leaf_name: leaf name
        :param str dst_ip: destination IP address
        :param str protocol: protocol type (choice: tcp, udp, icmp, ip)
        :param str filter_format: format: 'sport COMMA_SEPARATED_PORTS' or
                                  'dport COMMA_SEPARATED_PORTS'
        :param int rate_kbits: kilo bit
        :param int ceil_kbits: kilo bit
        :param int priority: new service priority

        :return: None
        :rtype: None
        """
        self.leaf_name = leaf_name
        self.dst_ip = dst_ip
        self.protocol = protocol
        self.filter_ = filter_format
        self.rate_kbits = rate_kbits
        self.ceil_kbits = ceil_kbits
        self.priority = priority



"""Get pi with rule by pi id API method."""
from ibsng.handler.handler import Handler


class getPIWithRuleByPIID(Handler):
    """Get pi with rule by pi id method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.pi_id, int)

    def setup(self, pi_id):
        """Setup required parameters.

        :param int pi_id: pi id

        :return: None
        :rtype: None
        """
        self.pi_id = pi_id



"""View bulk action API method."""
from ibsng.handler.handler import Handler


class viewBulkActionStatus(Handler):
    """View bulk action method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.action_id, int)

    def setup(self, action_id):
        """Setup required parameters.

        :param int action_id: bulk action id

        :return: None
        :rtype: None
        """
        self.action_id = action_id



"""Voucher change batch lock status API method."""
from ibsng.handler.handler import Handler


class voucherChangeBatchLockStatus(Handler):
    """Voucher change batch lock status method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.batch_id, int)
        self.is_valid(self.lock, bool)

    def setup(self, batch_id, lock):
        """Setup required parameters.

        :param int batch_id: batch id
        :param bool lock: lock

        :return: None
        :rtype: None
        """
        self.batch_id = batch_id
        self.lock = lock



"""Delete ip pool API method."""
from ibsng.handler.handler import Handler


class deleteIPpool(Handler):
    """Delete ip pool method class."""

    def control(self):
        """Validate inputs after method setup.

        :return: None
        :rtype: None
        """
        self.is_valid(self.ippool_name, str)

    def setup(self, ippool_name):
        """Setup required parameters.

        :param str ippool_name: ip pool name

        :return: None
        :rtype: None
        """
        self.ippool_name = ippool_name



"""Update group API method."""
from ibsng.handler.handler import Handler


class updateGroup(Handler):
    """Update group method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.group_id, int)
        self.is_valid(self.group_name, str)
        self.is_valid(self.status, str)
        self.is_valid(self.comment, str, False)

    def setup(self, group_id, group_name, status, comment=""):
        """Setup required parameters.

        :param int group_id: group id
        :param str group_name: new group name
        :param str status: status of group
        :param str comment: new group comment

        :return: None
        :rtype: None
        """
        self.group_id = group_id
        self.group_name = group_name
        self.status = status
        self.comment = comment



"""Get ISP of an admin AIP method."""
from ibsng.handler.handler import Handler


class getAdminISPID(Handler):
    """Get ISP ID of an admin class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.admin_username, str)

    def setup(self, admin_username):
        """Setup required parameters.

        :param str admin_username: admin authentication username

        :return: None
        :rtype: None
        """
        self.admin_username = admin_username



"""Update ip pool API method."""
from ibsng.handler.handler import Handler


class updateIPpool(Handler):
    """Update ip pool method class."""

    def control(self):
        """Validate inputs after method setup.

        :return: None
        :rtype: None
        """
        self.is_valid(self.ippool_id, int)
        self.is_valid(self.ippool_name, str)
        self.is_valid(self.comment, str)

    def setup(self, ippool_id, ippool_name, comment):
        """Setup required parameters.

        :param int ippool_id: ip pool id
        :param str ippool_name: ip pool name
        :param str comment: ip pool comment

        :return: None
        :rtype: None
        """
        self.ippool_id = ippool_id
        self.ippool_name = ippool_name
        self.comment = comment



"""Add new interface API method."""
from ibsng.handler.handler import Handler


class addInterface(Handler):
    """Add new interface class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.hostname, str)
        self.is_valid(self.interface_name, str)
        self.is_valid(self.host_type, str)
        self.is_valid_content(self.host_type, "(Linux|Mikrotik)")
        self.is_valid(self.comment, str, False)

    def setup(self, hostname, interface_name, host_type, comment=""):
        """Setup required parameters.

        :param str hostname: host name
        :param str interface_name: interface name
        :param str host_type: type of host (choice: Linux, Mikrotik)
        :param str comment: comment about this action

        :return: None
        :rtype: None
        """
        self.hostname = hostname
        self.interface_name = interface_name
        self.host_type = host_type
        self.comment = comment



"""Add invoice template API method."""
from ibsng.handler.handler import Handler


class addInvoiceTemplate(Handler):
    """Add invoice template method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.template_name, str)

    def setup(self, template_name):
        """Setup required parameters.

        :param str template_name: template name

        :return: None
        :rtype: None
        """
        self.template_name = template_name



"""Get ras ippools API method."""
from ibsng.handler.handler import Handler


class getRasIPpools(Handler):
    """Get ras ippools method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.ras_ip, str)
        self.is_valid_content(self.ras_ip, self.IP_PATTERN)

    def setup(self, ras_ip):
        """Setup required parameters.

        :param str ras_ip: ras ip address

        :return: None
        :rtype: None
        """
        self.ras_ip = ras_ip



"""Multi string get all API method."""
from ibsng.handler.handler import Handler


class multiStrGetAll(Handler):
    """Multi string get all method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.str_, str)
        self.is_valid(self.left_pad, bool)

    def setup(self, string, left_pad):
        """Setup required parameters.

        :param str str: string
        :param bool left_pad: left pad

        :return: None
        :rtype: None
        """
        self.str_ = string
        self.left_pad = left_pad



"""Set charge rule usage API method."""
from ibsng.handler.handler import Handler


class setOneChargeRuleUsage(Handler):
    """Set charge rule usage method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.user_id, int)
        self.is_valid(self.charge_rule_id, int)
        self.is_valid(self.credit_usage, int)
        self.is_valid(self.time_usage, int)
        self.is_valid(self.traffic_usage, int)

    def setup(self, user_id, charge_rule_id, credit_usage,
              time_usage, traffic_usage):
        """Setup required parameters.

        :param int user_id: ibsng user id
        :param int charge_rule_id: charge rule id
        :param int credit_usage: pass 0 if you want to reset
        :param int time_usage: in seconds, pass 0 if you want to reset
        :param int traffic_usage: in bytes, pass 0 if you want to reset

        :return: None
        :rtype: None
        """
        self.user_id = user_id
        self.charge_rule_id = charge_rule_id
        self.credit_usage = credit_usage
        self.time_usage = time_usage
        self.traffic_usage = traffic_usage



"""Generate OTP for an admin API method.

Generate one-time password for an admin.
"""
from ibsng.handler.handler import Handler


class generateOTP(Handler):
    """Generate OTP for an admin class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.pass_len, int)
        self.is_valid(self.pass_count, int)
        self.is_valid(self.admin_username, str)

    def setup(self, pass_len, pass_count, admin_username):
        """Setup required parameters.

        :param int pass_len: password length
        :param int pass_count: count of password
        :param str admin_username: admin authentication username

        :return: None
        :rtype: None
        """
        self.pass_len = pass_len
        self.pass_count = pass_count
        self.admin_username = admin_username



"""Delete leaf service API method."""
from ibsng.handler.handler import Handler


class delLeafService(Handler):
    """Delete leaf service class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.leaf_service_id, self.ID_PATTERN)
        self.is_valid(self.leaf_name, str)

    def setup(self, leaf_service_id, leaf_name):
        """Setup required parameters.

        :param int leaf_service_id: leaf service ID
        :param str leaf_name: leaf name

        :return: None
        :rtype: None
        """
        self.leaf_service_id = leaf_service_id
        self.leaf_name = leaf_name



"""Delete group API method."""
from ibsng.handler.handler import Handler


class delGroup(Handler):
    """Delete group method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.group_name, str)

    def setup(self, group_name):
        """Setup required parameters.

        :param str group_name: group name

        :return: None
        :rtype: None
        """
        self.group_name = group_name



"""Delete user API method."""
from ibsng.handler.handler import Handler


class delUser(Handler):
    """Delete user method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.user_id, str)
        self.is_valid(self.del_connection_logs, bool)
        self.is_valid(self.del_audit_logs, bool)
        self.is_valid(self.delete_comment, str, False)

    def setup(self, user_ids, del_connection_logs,
              del_audit_logs, delete_comment=""):
        """Setup required parameters.

        :param list user_id: list of ibsng user ids
        :param str delete_comment: comment for this action
        :param bool del_connection_logs: pass true if you want users'
                                         connection logs to be deleted too
        :param bool del_audit_logs: pass true if you want users' audit logs to
                                    be deleted too

        :return: None
        :rtype: None
        """
        self.user_id = ",".join(map(str, user_ids))
        self.del_connection_logs = del_connection_logs
        self.del_audit_logs = del_audit_logs
        self.delete_comment = delete_comment



"""Set LDAP search options by domain API method."""
from ibsng.handler.handler import Handler


class setLDAPSearchOptionsByDomain(Handler):
    """Set LDAP search options by domain method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.domain, str)
        self.is_valid(self.search_filter, str)
        self.is_valid(self.search_base, str)

    def setup(self, domain, search_filter, search_base):
        """Setup required parameters.

        :param str domain: name of ldap domain to be queried
        :param str search_filter: ldap query search filter by which ldap users
                                  will be fetched
        :param str search_base: ldap query search base by which ldap users
                                will be fetched

        :return: None
        :rtype: None
        """
        self.domain = domain
        self.search_filter = search_filter
        self.search_base = search_base



"""Get charge info API method."""
from ibsng.handler.handler import Handler


class getChargeInfo(Handler):
    """Get charge info method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        if getattr(self, "charge_name"):
            self.is_valid(self.charge_name, str)
        elif getattr(self, "charge_id"):
            self.is_valid(self.charge_id, int)

    def setup(self, charge_name, charge_id=None):
        """Setup required parameters.

        :param charge_name: charge name
        :type charge_name: str
        :param charge_id: charge id
        :type charge_id: int

        :return: None
        :rtype: None
        """
        if charge_name:
            self.charge_name = charge_name
        elif charge_id:
            self.charge_id = charge_id



"""Get invoice profile by name API method."""
from ibsng.handler.handler import Handler


class getInvoiceProfileByName(Handler):
    """Get invoice profile by name method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.profile_name, str)

    def setup(self, profile_name):
        """Setup required parameters.

        :param str profile_name: ibsng profile name

        :return: None
        :rtype: None
        """
        self.profile_name = profile_name



"""Get all template names API method."""
from ibsng.handler.handler import Handler


class getAllTemplateNames(Handler):
    """Get all template names method class."""

    def setup(self, **kwargs):
        """Setup required parameters.

        :param dict kwargs: input args

        :return: void
        :rtype: void
        """
        for key, value in kwargs.items():
            setattr(self, key, value)



"""Invoice put user on dept API method."""
from ibsng.handler.handler import Handler


class invoicePutUserOnDept(Handler):
    """Invoice put user on dept method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.pi_id_list, list)

    def setup(self, pi_id_list):
        """Setup required parameters.

        :param list pi_id_list: list of proforma invoice ids

        :return: None
        :rtype: None
        """
        self.pi_id_list = pi_id_list



"""Get auth data API method."""
from ibsng.handler.handler import Handler


class getAuthData(Handler):
    """Get auth data method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.auth_session, int)
        self.is_valid_content(self.auth_session, self.ID_PATTERN)

    def setup(self, session_id):
        """Setup required parameters.

        :param str auth_session: session id

        :return: None
        :rtype: None
        """
        self.auth_session = session_id



"""Delete custom field API method."""
from ibsng.handler.handler import Handler


class deleteCustomField(Handler):
    """Delete custom field method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.name, str)

    def setup(self, name):
        """Setup required parameters.

        :param str name: custom field name

        :return: None
        :rtype: None
        """
        self.name = name



"""Change deposit API method."""
from ibsng.handler.handler import Handler


class changeDeposit(Handler):
    """Change deposit method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.user_id, int)
        self.is_valid(self.deposit, float)
        self.is_valid(self.is_absolute_change, bool)
        self.is_valid(self.deposit_comment, str, False)

    def setup(self, user_id, deposit, is_absolute_change, deposit_comment=""):
        """Setup required parameters.

        :param str user_id: ibsng user id
        :param float deposit: new deposit
        :param bool is_absolute_change: 'false' for relative change (add to
                                         user's deposit), 'true' for absolute
                                         change (set user's deposit)
        :param str deposit_comment: comment for this action

        :return: None
        :rtype: None
        """
        self.user_id = user_id
        self.deposit = deposit
        self.is_absolute_change = is_absolute_change
        self.deposit_comment = deposit_comment



"""IBSng main access file."""
from .connection.connection import Connection
IBSng = Connection
MAJOR_VERSION = '1'
MINOR_VERSION = '2'
MICRO_VERSION = '2'
VERSION = "{}.{}.{}".format(MAJOR_VERSION, MINOR_VERSION, MICRO_VERSION)



"""Update invoice profile API method."""
from ibsng.handler.handler import Handler


class updateInvoiceProfile(Handler):
    """Update invoice profile method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.profile_id, int)
        self.is_valid(self.profile_name, str)
        self.is_valid(self.isp_name, str, False)
        self.is_valid(self.rules, list, False)
        self.is_valid(self.comment, str, False)

    def setup(self, profile_id, profile_name,
              isp_name="", rules=[], comment=""):
        """Setup required parameters.

        :param int profile_id: profile id
        :param str profile_name: new profile name
        :param str isp_name: new isp name
        :param list rules: new rules
        :param str comment: new comment

        :return: None
        :rtype: None
        """
        self.profile_id = profile_id
        self.profile_name = profile_name
        self.isp_name = isp_name
        self.rules = rules
        self.comment = comment



"""Save search user API method."""
from ibsng.handler.handler import Handler


class saveSearchUser(Handler):
    """Save search users method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.conds, dict)
        self.is_valid(self.basic, list)
        self.is_valid(self.attrs, list)
        self.is_valid(self.order_by, str)
        self.is_valid(self.desc, bool)
        self.is_valid(self.output_type, str)

    def setup(self, conds, basic, attrs, order_by, desc, output_type):
        """Setup required parameters.

        :param dict conds: conditions
        :param list basic: basic user attrs.
                           (user_id, creation_date, credit, deposit, group_id,
                           group_name, isp_id, isp_name, nearest_exp_date,
                           recharge_deposit, status)
        :param list attrs: output of getUserInfo for complete list of attrs.
                           (normal_username, normal_password, first_login,
                           real_first_login, nearest_exp_date,
                           nearest_exp_date_epoch, time_to_nearest_exp_date,
                           charge_rule_usage, charge, last_renew,
                           abs_exp_date, assign_dns, assign_dns_1,
                           assign_dns_2, assign_ip, assign_netmask,
                           assign_route_ip, caller_id, fast_dial,
                           forward_number, asterisk_extension,
                           asterisk_ras_ip, asterisk_ras_id,
                           asterisk_ext_details, asterisk_sip_details,
                           asterisk_iax_details, asterisk_did_ras_ip,
                           asterisk_did_ras_id, comment, name, email, phone,
                           cell_phone, birthdate, address, postal_code,
                           credit_limit, exp_date_temp_extend,
                           temp_extend_log_id, exp_from_creation_date,
                           exp_from_first_login, exp_from_renew, exp_monthly,
                           extra_charge_profile, fast_dial, forward_number,
                           idle_threshold_timespan, idle_bytes_threshold,
                           idle_timeout, bw_send_rate, bw_receive_rate
                           in_ras_bw, in_ras_ippool, ippool, limit_caller_id,
                           limit_mac, limit_port, limit_ras, limit_ras,
                           limit_station_ip, limit_usage_days, lock,
                           assigned_endpoint, mc_charge, mc_charge_send_latin,
                           mc_charge_send_unicode, mc_charge_recv,
                           mc_min_user_credit, email_address, mail_quota,
                           mail_usage, mikrotik_address_list, mikrotik_bw,
                           mikrotik_bw_recv_rate, mikrotik_bw_send_rate,
                           mikrotik_bw_recv_burst, mikrotik_bw_send_burst,
                           mikrotik_bw_recv_burst_threshold,
                           mikrotik_bw_send_burst_threshold,
                           mikrotik_bw_recv_burst_time,
                           mikrotik_bw_send_burst_time,
                           mikrotik_bw_priority,
                           mikrotik_bw_min_recv_rate,
                           mikrotik_bw_min_send_rate,
                           multi_login,
                           notification_profile,
                           time_periodic_accounting_monthly,
                           time_periodic_accounting_monthly_limit,
                           time_periodic_accounting_monthly_usage,
                           time_periodic_accounting_monthly_reset,
                           traffic_periodic_accounting_monthly,
                           traffic_periodic_accounting_monthly_limit,
                           traffic_periodic_accounting_monthly_usage,
                           traffic_periodic_accounting_monthly_reset,
                           time_periodic_accounting_daily,
                           time_periodic_accounting_daily_limit,
                           time_periodic_accounting_daily_usage,
                           time_periodic_accounting_daily_reset,
                           traffic_periodic_accounting_daily,
                           traffic_periodic_accounting_daily_limit,
                           traffic_periodic_accounting_daily_usage,
                           traffic_periodic_accounting_daily_reset,
                           time_periodic_accounting_weekly,
                           time_periodic_accounting_weekly_limit,
                           time_periodic_accounting_weekly_usage,
                           time_periodic_accounting_weekly_reset,
                           traffic_periodic_accounting_weekly,
                           traffic_periodic_accounting_weekly_limit,
                           traffic_periodic_accounting_weekly_usage,
                           traffic_periodic_accounting_weekly_reset,
                           monthly_periodic_credit_change,
                           monthly_periodic_credit_change_day,
                           monthly_periodic_credit_change_type,
                           monthly_periodic_credit_change_amount,
                           monthly_periodic_credit_change_after_first_login,
                           monthly_periodic_credit_change_negate_credit,
                           monthly_periodic_credit_change_do_on_first_login,
                           monthly_periodic_credit_change_reset,
                           monthly_periodic_credit_change_reset,
                           daily_periodic_credit_change,
                           daily_periodic_credit_change_type,
                           daily_periodic_credit_change_amount,
                           daily_periodic_credit_change_after_first_login,
                           daily_periodic_credit_change_negate_credit,
                           daily_periodic_credit_change_do_on_first_login,
                           daily_periodic_credit_change_reset,
                           persistent_lan_ip, persistent_lan_mac,
                           persistent_lan_ras_ip, plan, radius_attrs,
                           rel_exp_date, rel_exp_date_unit,
                           rel_exp_date_negative_credit,
                           rel_exp_date_negative_credit_unit,
                           renew_next_group, second_normal_username,
                           serial, session_timeout, voip_preferred_language,
                           voip_username, negative_credit_start,
                           save_bw_usage, allow_user_disconnect,
                           deny_change_password, auto_renew,
                           reset_negative_credit_on_renew,
                           renew_do_not_reset_credit,
                           renew_do_not_reset_credit_on_state_package,
                           renew_do_not_reset_credit_on_state_recharged,
                           auto_renew_on_credit_finish
                           auto_renew_do_not_reset_credit
                           auto_recharge, allow_recharge_by_voucher,
                           prevent_web_login, enable_failed_login
                           ignore_idle_threshold_static_ip,
                           ignore_idle_threshold,
                           caller_id_bind_on_login,
                           limit_mac_bind_on_login,
                           limit_port_ibind_on_login,
                           limit_ras_bind_on_login,
                           limit_station_ip_bind_on_login,
                           limit_caller_id_allow_not_defined,
                           is_deposit_transferee,
                           is_deposit_transferer)
        :param str order_by: order by a field (normal_username,
                             voip_username, user_id, group_id,
                             creation_date, owner_id, credit, deposit,
                             first_login, serial, serial_numeric)
        :param bool desc: descending
        :param str output_type: output type (csv, pdf)

        :return: None
        :rtype: None
        """
        self.conds = conds
        self.basic = basic
        self.attrs = attrs
        self.order_by = order_by
        self.desc = desc
        self.output_type = output_type



"""Update bandwidth static IP API method."""
from ibsng.handler.handler import Handler


class updateBwStaticIP(Handler):
    """Update bandwidth static IP class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.static_ip_id, int)
        self.is_valid_content(self.static_ip_id, self.ID_PATTERN)
        self.is_valid(self.ip_addr, str)
        self.is_valid_content(self.ip_addr, self.IP_PATTERN)
        self.is_valid(self.tx_leaf_name, str)
        self.is_valid(self.rx_leaf_name, str)

    def setup(self, static_ip_id, ip_addr, tx_leaf_name, rx_leaf_name):
        """Setup required parameters.

        :param int static_ip_id: static ip id
        :param str ip_addr: new ip address
        :param str tx_leaf_name: new leaf name
        :param str rx_leaf_name: new leaf name

        :return: None
        :rtype: None
        """
        self.static_ip_id = static_ip_id
        self.ip_addr = ip_addr
        self.tx_leaf_name = tx_leaf_name
        self.rx_leaf_name = rx_leaf_name



"""Search batch API method."""
from ibsng.handler.handler import Handler


class searchBatch(Handler):
    """Search batch method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.conds, dict)
        self.is_valid(self.index_from, int)
        self.is_valid(self.index_to, int)
        self.is_valid(self.sort_by, str)
        self.is_valid(self.desc, bool)

    def setup(self, conds, index_from, index_to, sort_by, desc):
        """Setup required parameters.

        :param dict conds: conditions
        :param int index_from: from index
        :param int index_to: end index
        :param str sort_by: sort result by specific field (batch_id,
                            batch_name, batch_credit)
        :param bool desc: descending

        :return: None
        :rtype: None
        """
        self.conds = conds
        self.index_from = index_from
        self.index_to = index_to
        self.sort_by = sort_by
        self.desc = desc



"""Update user attrs API method."""
from ibsng.handler.handler import Handler


class updateUserAttrs(Handler):
    """Update user attrs method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.user_id, int)
        self.is_valid(self.attrs, dict)
        self.is_valid(self.to_del_attrs, list)

    def setup(self, user_id, attrs, to_del_attrs):
        """Setup required parameters.

        :param str user_id: ibsng user id
        :param dict attrs: attributes with values
        :param list to_del_attrs: list of attributes to be deleted

        :return: None
        :rtype: None
        """
        self.user_id = user_id
        self.attrs = attrs
        self.to_del_attrs = to_del_attrs



"""Change user credit API method."""
from ibsng.handler.handler import Handler


class changeCredit(Handler):
    """Change user credit method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.user_id, int)
        self.is_valid(self.credit, float)
        self.is_valid(self.is_absolute_change, bool)
        self.is_valid(self.credit_comment, str, False)

    def setup(self, user_id, credit, is_absolute_change, credit_comment=""):
        """Setup required parameters.

        :param str user_id: ibsng user id
        :param float credit: new credit value
        :param bool is_absolute_change: 'false' for relative change (add to
                                        user's credit), 'true' for absolute
                                        change (set user's credit)
        :param str credit_comment: comment for this credit

        :return: None
        :rtype: None
        """
        self.user_id = user_id
        self.credit = credit
        self.is_absolute_change = is_absolute_change
        self.credit_comment = credit_comment



"""Bulk action to update user attrs API method."""
from ibsng.handler.handler import Handler


class bulkUpdateUserAttrs(Handler):
    """Bulk action to update user attrs method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.conds, dict)
        self.is_valid(self.attrs_to_update, dict, False)
        self.is_valid(self.attrs_to_del, list, False)

    def setup(self, conds, attrs_to_update={}, attrs_to_del=[]):
        """Setup required parameters.

        :param dict conds: conditions
        :param dict attrs_to_update: attributes with new values
        :param list attrs_to_del: list of attributes which should be deleted

        :return: None
        :rtype: None
        """
        self.conds = conds
        self.attrs_to_update = attrs_to_update
        self.attrs_to_del = attrs_to_del



"""Get real host interfaces API method."""
from ibsng.handler.handler import Handler


class getRealHostInterfaces(Handler):
    """Get real host interfaces class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.hostname, str)
        self.is_valid(self.host_type, str)
        self.is_valid_content(self.host_type, "(Linux|Mikrotik)")

    def setup(self, hostname, host_type):
        """Setup required parameters.

        :param str hostname: host name
        :param str host_type: host type(Linux, Mikrotik)

        :return: None
        :rtype: None
        """
        self.hostname = hostname
        self.host_type = host_type



"""Change ISP deposit API method."""
from ibsng.handler.handler import Handler


class changeISPDeposit(Handler):
    """Change ISP deposit method class."""

    def control(self):
        """Validate inputs after method setup.

        :rtype: None
        :return: None
        """
        self.is_valid(self.isp_name, str)
        self.is_valid(self.deposit_amount, float)
        self.is_valid(self.comment, str)

    def setup(self, isp_name, deposit_amount, comment=""):
        """Setup required parameters.

        :param str isp_name: isp name
        :param float deposit_amount: deposit amount
        :param str comment: comment

        :return: None
        :rtype: None
        """
        self.isp_name = isp_name
        self.deposit_amount = deposit_amount
        self.comment = comment



"""Set Feshfeshe params API method."""
from ibsng.handler.handler import Handler


class setFeshfesheParams(Handler):
    """Set Feshfeshe params method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.user_id, int)
        self.is_valid(self.traffic, float)
        self.is_valid(self.date_from, str)
        self.is_valid(self.date_to, str)

    def setup(self, user_id, traffic, date_from, date_to):
        """Setup required parameters.

        :param int user_id: ibsng user id
        :param int traffic: Traffic, in Mega Bytes
        :param str_datetime date_from: start date and time,
                                       format '%Y-%m-%d %H:%M'
        :param str_datetime date_to: end date and time,
                                     format '%Y-%m-%d %H:%M'

        :return: None
        :rtype: None
        """
        self.user_id = user_id
        self.traffic = traffic
        self.date_from = date_from
        self.date_to = date_to



"""Update leaf service API method."""
from ibsng.handler.handler import Handler


class updateLeafService(Handler):
    """Update leaf service class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.leaf_name, str)
        self.is_valid(self.leaf_service_id, int)
        self.is_valid_content(self.leaf_service_id, self.ID_PATTERN)
        self.is_valid(self.dst_ip, str)
        self.is_valid_content(self.dst_ip, self.IP_PATTERN)
        self.is_valid(self.leaf_name, str)
        self.is_valid(self.protocol, str)
        self.is_valid_content(self.protocol, "(tcp|udp|icmp|ip)")
        self.is_valid(self.filter_, str)
        self.is_valid(self.rate_kbits, int)
        self.is_valid(self.ceil_kbits, int)
        self.is_valid(self.priority, int)
        self.is_valid_content(self.priority, self.POSITIVE_NUMBER)

    def setup(self, leaf_name, leaf_service_id, dst_ip, protocol,
              filter_format, rate_kbits, ceil_kbits, priority):
        """Setup required parameters.

        :param str leaf_name: leaf name
        :param int leaf_service_id: leaf service ID
        :param str dst_ip: new destination IP
        :param str protocol: new protocol (choice: tcp, udp, icmp, ip)
        :param str filter_format: format: 'sport COMMA_SEPARATED_PORTS' or
                                  'dport COMMA_SEPARATED_PORTS'
        :param int rate_kbits: new kilo bits
        :param int ceil_kbits: new kilo bits
        :param int priority: new priority

        :return: None
        :rtype: None
        """
        self.leaf_name = leaf_name
        self.leaf_service_id = leaf_service_id
        self.dst_ip = dst_ip
        self.protocol = protocol
        self.filter_ = filter_format
        self.rate_kbits = rate_kbits
        self.ceil_kbits = ceil_kbits
        self.priority = priority



"""Update user comments API method."""
from ibsng.handler.handler import Handler


class updateUserComments(Handler):
    """Update user comments method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.update_dic, dict)

    def setup(self, update_dict):
        """Setup required parameters.

        :param dict update_dict: key value attributes

        :return: None
        :rtype: None
        """
        self.update_dic = update_dict



"""Get invoice profile by userid API method."""
from ibsng.handler.handler import Handler


class getInvoiceProfileByUserID(Handler):
    """Get invoice profile by user id method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.user_id, int)

    def setup(self, user_id):
        """Setup required parameters.

        :param int user_id: ibsng user id

        :return: None
        :rtype: None
        """
        self.user_id = user_id



"""Update load balancing ip pool API method."""
from ibsng.handler.handler import Handler


class updateLoadBalancingIPpool(Handler):
    """Update load balancing ip pool method class."""

    def control(self):
        """Validate inputs after method setup.

        :return: None
        :rtype: None
        """
        self.is_valid(self.ippool_id, int)
        self.is_valid(self.ippool_name, str)
        self.is_valid(self.ippool_comment, str)
        self.is_valid_choice(self.balancing_strategy, ['distributive',
                                                       'fill_first'])

    def setup(self, ippool_id, ippool_name, ippool_comment,
              balancing_strategy, children_ippool_percentages):
        """Setup required parameters.

        :param int ippool_id: ip pool id
        :param str ippool_name: ip pool name
        :param str ippool_comment: ip pool comment
        :param choice balancing_strategy: balancing strategy
        :param dict children_ippool_percentages: keys are IPPool Names, values
                                                 are percentage numbers

        :return: None
        :rtype: None
        """
        self.ippool_id = ippool_id
        self.ippool_name = ippool_name
        self.ippool_comment = ippool_comment
        self.balancing_strategy = balancing_strategy
        self.children_ippool_percentages = children_ippool_percentages



"""Add bandwidth static IP API method."""
from ibsng.handler.handler import Handler


class addBwStaticIP(Handler):
    """Add bandwidth static IP class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.ip_addr, str)
        self.is_valid(self.tx_leaf_name, str)
        self.is_valid(self.rx_leaf_name, str)

    def setup(self, ip_addr, tx_leaf_name, rx_leaf_name):
        """Setup required parameters.

        :param str ip_addr: IP address
        :param str tx_leaf_name: leaf name
        :param str rx_leaf_name: leaf name

        :return: None
        :rtype: None
        """
        self.ip_addr = ip_addr
        self.tx_leaf_name = tx_leaf_name
        self.rx_leaf_name = rx_leaf_name



"""IBSng 'general' exceptions handler."""
from .base import IBSngException


class ChainFormatError(IBSngException):
    """Raise when method or handler called in invalid chain format."""

    _err_code = 201

    def __init__(self, chain_list):
        """Raise when method or handler called in invalid chain format.

        :param chain_list: list of all chain items
        :type chain_list: list
        """
        message = "Chain format '{}' is incorrect format.".format(chain_list)
        super(ChainFormatError, self).__init__(message)



"""Get node info API method."""
from ibsng.handler.handler import Handler


class getNodeInfo(Handler):
    """Get node info class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.node_id, int)
        self.is_valid_content(self.node_id, self.ID_PATTERN)

    def setup(self, node_id):
        """Setup required parameters.

        :param int node_id: node id

        :return: None
        :rtype: None
        """
        self.node_id = node_id



"""List groups API method."""
from ibsng.handler.handler import Handler


class listGroups(Handler):
    """List groups method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.isp_id, str)
        self.is_valid_content(self.isp_id, self.IDS_PATTERN)
        self.is_valid(self.active_only, bool)

    def setup(self, isp_ids, active_only=False):
        """Setup required parameters.

        :param str isp_ids: comma-seprated list of isp ids
        :param bool active_only: pass true if you only want active groups

        :return: None
        :rtype: None
        """
        self.active_only = active_only
        self.isp_id = ",".join(map(str, isp_ids))



"""Delete permission template API method."""
from ibsng.handler.handler import Handler


class deletePermTemplate(Handler):
    """Delete permission template method class."""

    def control(self):
        """Validate inputs after method setup.

        :return: None
        :rtype: None
        """
        self.is_valid(self.perm_template_name, str)

    def setup(self, perm_template_name):
        """Setup required parameters.

        :param str perm_template_name: permission template name

        :return: None
        :rtype: None
        """
        self.perm_template_name = perm_template_name



"""Get remain duration and bytes API method."""
from ibsng.handler.handler import Handler


class getRemainingDurationAndBytes(Handler):
    """Get remain duration and bytes method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        if hasattr(self, "user_id"):
            self.is_valid(self.user_id, int)

    def setup(self, user_id=None):
        """Setup required parameters.

        :param int user_id: ibsng user id

        :return: None
        :rtype: None
        """
        if user_id:
            self.user_id = user_id



"""IBSng 'unknown' exceptions handler."""
from .base import IBSngException


class UnknownError(IBSngException):
    """Raise when method result contains unknown error."""

    _err_code = 601

    def __init__(self, error_text):
        """Raise when method result contains unknown error.

        :param error_text: method error body
        :type error_text: str
        """
        message = "Server method returned unknown error: {}"
        message = message.format(error_text)
        super(UnknownError, self).__init__(message)



"""Get tree API method."""
from ibsng.handler.handler import Handler


class getTree(Handler):
    """Get tree class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.hostname, str)
        self.is_valid(self.interface_name, str)

    def setup(self, hostname, interface_name):
        """Setup required parameters.

        :param str hostname: host name
        :param str interface_name: interface name

        :return: None
        :rtype: None
        """
        self.hostname = hostname
        self.interface_name = interface_name



"""Get all username of admins API method."""
from ibsng.handler.handler import Handler


class getAllAdminUsernames(Handler):
    """Get all username of admins class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        if getattr(self, "isp_name"):
            self.is_valid(self.isp_name, str)

    def setup(self, isp_name=None):
        """Setup required parameters.

        :param str isp_name: isp name (optional)

        :return: None
        :rtype: None
        """
        if isp_name:
            self.isp_name = isp_name



"""Expire session API method."""
from ibsng.handler.handler import Handler


class expireSession(Handler):
    """Expire session method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.session_id, int)
        self.is_valid_content(self.session_id, self.ID_PATTERN)

    def setup(self, session_id):
        """Setup required parameters.

        :param str session_id: session id

        :return: None
        :rtype: None
        """
        self.session_id = session_id



"""Get user info API method."""
from ibsng.handler.handler import Handler


class getUserInfo(Handler):
    """Get user info class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        if hasattr(self, "user_id"):
            self.is_valid(self.user_id, int)
        if hasattr(self, "normal_username"):
            self.is_valid(self.normal_username, str)
        if hasattr(self, "voip_username"):
            self.is_valid(self.voip_username, str)
        if hasattr(self, "serial"):
            self.is_valid(self.serial, str)
        if hasattr(self, "phone"):
            self.is_valid(self.phone, str)

    def setup(self, user_id=None, normal_username=None,
              voip_username=None, serial=None, phone=None):
        """Setup required parameters.

        :param int user_id: ibsng user id
        :param str normal_username: authentication username
        :param str voip_username: voip username
        :param str serial: serial code
        :param str phone: phone number

        :return: None
        :rtype: None
        """
        if user_id:
            self.user_id = user_id
        elif normal_username:
            self.normal_username = normal_username
        elif voip_username:
            self.voip_username = voip_username
        elif serial:
            self.serial = serial
        elif phone:
            self.phone = phone



"""Voucher add new user API method."""
from ibsng.handler.handler import Handler


class voucherAddNewUser(Handler):
    """Voucher add new user method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.voucher_pin, str)
        self.is_valid(self.isp_name, str)
        self.is_valid(self.group_name, str)

    def setup(self, voucher_pin, isp_name, group_name):
        """Setup required parameters.

        :param str voucher_pin: voucher pin
        :param str isp_name: isp name
        :param str group_name: group name

        :return: None
        :rtype: None
        """
        self.voucher_pin = voucher_pin
        self.isp_name = isp_name
        self.group_name = group_name



import unittest
from ibsng.util import string


class TestUtil(unittest.TestCase):
    def test_connection_string_parser(self):
        r1 = string. \
            connection_string_parser("user:pass@http://192.168.1.1:8080")
        self.assertIsInstance(r1, tuple)
        self.assertEqual(r1[1], "http://192.168.1.1:8080")
        self.assertIsInstance(r1[0], dict)

        with self.assertRaises(Exception) as context:
            string.connection_string_parser("user:pass@192.168.1.1:8080")
        self.assertIn("Invalid connection string", str(context.exception))

        with self.assertRaises(Exception) as context:
            string.connection_string_parser("something wrong")
        self.assertIn("Invalid connection string", str(context.exception))

    def test_camel_to_snake(self):
        r1 = string.camel_to_snake("ibsngTest")
        self.assertEqual(r1, "ibsng_test")

        r1 = string.camel_to_snake("IbsngTest")
        self.assertEqual(r1, "ibsng_test")

        r1 = string.camel_to_snake("323232")
        self.assertEqual(r1, "323232")

        r1 = string.camel_to_snake("Ibsng_test")
        self.assertEqual(r1, "ibsng_test")
 
        r1 = string.camel_to_snake("_Ibsng_test")
        self.assertEqual(r1, "_ibsng_test")

if __name__ == "__main__":
    unittest.main()



"""Get users by phone API method."""
from ibsng.handler.handler import Handler


class getUsersWithPhone(Handler):
    """Get users by phone method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.phone, str)

    def setup(self, phone):
        """Setup required parameters.

        :param str phone: ibsng users phone

        :return: None
        :rtype: None
        """
        self.phone = phone



"""Get system notification API method."""
from ibsng.handler.handler import Handler


class getNotifications(Handler):
    """Get system notification method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.last_notifications, int)
        self.is_valid_content(self.last_notifications, self.POSITIVE_NUMBER)
        self.is_valid(self.only_unread, bool)

    def setup(self, last_notifications=5, only_unread=False):
        """Setup required parameters.

        :param int last_notifications: last notifications (default is 5)
        :param bool only_unread: just get unread notification

        :return: None
        :rtype: None
        """
        self.last_notifications = last_notifications
        self.only_unread = only_unread



"""Get LDAP IBSng mapping API method."""
from ibsng.handler.handler import Handler


class getLDAPIBSMapping(Handler):
    """Get LDAP IBSng mapping method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.domain, str)

    def setup(self, domain_name):
        """Setup required parameters.

        :param str domain_name: name of ldap domain

        :return: None
        :rtype: None
        """
        self.domain = domain_name



"""Update admin info API method."""
from ibsng.handler.handler import Handler


class updateAdminInfo(Handler):
    """Update admin info class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.admin_id, int)
        self.is_valid(self.admin_username, str)
        self.is_valid(self.admin_isp_name, str)
        self.is_valid(self.admin_locked, bool)
        self.is_valid(self.name, str)
        self.is_valid(self.email, str)
        self.is_valid_content(self.email, self.EMAIL_PATTERN)
        self.is_valid(self.comment, str, False)
        self.is_valid(self.admin_has_otp, bool)
        self.is_valid(self.admin_request_limit, int)

    def setup(self, admin_id, admin_username, admin_isp_name, admin_locked,
              name, email, admin_has_otp, admin_request_limit, comment=""):
        """Setup required parameters.

        :param int admin_id: admin id
        :param str admin_username: admin authentication username
        :param str admin_isp_name: admin isp name
        :param bool admin_locked: lock admin
        :param name name: admin real name
        :param name email: admin email address
        :param bool admin_has_otp: has admin otp
        :param int admin_request_limit: limitation on admin requests
        :param name comment: comment about this action

        :return: None
        :rtype: None
        """
        self.admin_id = admin_id
        self.admin_username = admin_username
        self.admin_isp_name = admin_isp_name
        self.admin_locked = admin_locked
        self.name = name
        self.email = email
        self.admin_has_otp = admin_has_otp
        self.admin_request_limit = admin_request_limit
        self.comment = comment



"""Delete ISP API method."""
from ibsng.handler.handler import Handler


class deleteISP(Handler):
    """Delete ISP method class."""

    def control(self):
        """Validate inputs after method setup.

        :return: None
        :rtype: None
        """
        self.is_valid(self.isp_name, str)

    def setup(self, isp_name):
        """Setup required parameters.

        :param str isp_name: isp name

        :return: None
        :rtype: None
        """
        self.isp_name = isp_name



"""Voucher and batch API method."""
from ibsng.handler.handler import Handler


class voucherAddBatch(Handler):
    """Voucher and batch method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.batch_dict, dict)
        self.is_valid(self.pin_prefix, str)
        self.is_valid(self.pin_len, int)
        self.is_valid(self.serial_prefix, str)
        self.is_valid(self.serial_start, int)
        self.is_valid(self.count, int)

    def setup(self, batch_dict, pin_prefix, pin_len,
              serial_prefix, serial_start, count):
        """Setup required parameters.

        :param dict batch_dict: batch dictionary
        :param str pin_prefix: pin prefix
        :param int pin_len: pin length
        :param str serial_prefix: serial prefix
        :param int serial_start: serial start
        :param int count: count

        :return: None
        :rtype: None
        """
        self.batch_dict = batch_dict
        self.pin_prefix = pin_prefix
        self.pin_len = pin_len
        self.serial_prefix = serial_prefix
        self.serial_start = serial_start
        self.count = count



"""Get users by cellphone API method."""
from ibsng.handler.handler import Handler


class getUsersWithCellPhone(Handler):
    """Get users by cellphone method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.cell_phone, str)

    def setup(self, cellphone):
        """Setup required parameters.

        :param str cellphone: ibsng users cellphone

        :return: None
        :rtype: None
        """
        self.cell_phone = cellphone



"""IBSng base exceptions handler.

Exception numbers (starts with):
    1: connections
    2: generals
    3: handlers
    4: methods
    5: authentication
    6: unknown
"""
import datetime


class IBSngException(Exception):
    """Base IBSng exception handler."""

    timestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    _err_code = None

    def __init__(self, message):
        """Base IBSng exception handler.

        :param message: exception message
        :type message: str
        """
        self._err_msg = message
        message = "[{}]:{} - {}".format(self.timestamp,
                                        message,
                                        self._err_code)
        super(IBSngException, self).__init__(message)

    @property
    def error_code(self):
        """Return exception error code.

        :return: exception error code
        :rtype: int
        """
        return self._err_code

    @property
    def error_msg(self):
        """Return exception error message.

        :return: exception error message
        :rtype: str
        """
        return self._err_msg



"""Kill user API method."""
from ibsng.handler.handler import Handler


class killUser(Handler):
    """Kill user method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.user_id, str)
        self.is_valid(self.ras_ip, str)
        self.is_valid(self.unique_id_val, str)
        self.is_valid(self.kill, bool)

    def setup(self, user_ids, ras_ips, unique_id_val, kill=False):
        """Setup required parameters.

        :param list user_id: ibsng user ids
        :param list ras_ips: ras ips
        :param str unique_id_val: unique id value (for exmaple port, caller id,
                                  etc)
        :param bool kill: true means 'Kill User', false means 'Clear User'

        :return: None
        :rtype: None
        """
        self.user_id = ",".join(map(str, user_ids))
        self.ras_ip = ",".join(map(str, ras_ips))
        self.unique_id_val = unique_id_val
        self.kill = kill



"""Update custom field API method."""
from ibsng.handler.handler import Handler


class updateCustomField(Handler):
    """Update custom field method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.custom_field_id, int)
        self.is_valid(self.name, str)
        self.is_valid(self.description, str, False)
        self.is_valid(self.value_type, str)
        self.is_valid_content(self.value_type, "(string|integer|float)")
        self.is_valid(self.interface_type, str)
        self.is_valid_content(self.interface_type, "(text_field|single_select\
                                                   |radio_button|checkbox)")
        self.is_valid(self.allowable_values, list, False)
        self.is_valid(self.mandatory, bool)
        self.is_valid(self.comment, str, False)

    def setup(self, custom_field_id, name, description, value_type,
              interface_type, allowable_values, mandatory, comment=""):
        """Setup required parameters.

        :param int custom_field_id: ibsng custom field id
        :param str name: custom field name
        :param str description: custom field description
        :param str value_type: custom field value type: string, integer, float
        :param str interface_type: custom field interface type: text_field,
                                   single_select, radio_button, checkbox
        :param list allowable_values: type of items depends on value_type param
        :param bool mandatory: mark field as required
        :param str comment: custom field comment

        :return: None
        :rtype: None
        """
        self.custom_field_id = custom_field_id
        self.name = name
        self.description = description
        self.value_type = value_type
        self.interface_type = interface_type
        self.allowable_values = allowable_values
        self.mandatory = mandatory
        self.comment = comment



"""Bulk change user API method."""
from ibsng.handler.handler import Handler


class bulkChangeUserDeposit(Handler):
    """Bulk change user method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.conds, dict)
        self.is_valid(self.deposit, float)
        self.is_valid(self.change_type, str)
        self.is_valid_content(self.change_type, ("ADD|SET|MULTIPLY"))
        self.is_valid(self.deposit_comment, str, False)

    def setup(self, conds, deposit, change_type, deposit_comment=""):
        """Setup required parameters.

        :param dict conds: conditions
        :param float deposit: initialize deposit
        :param str change_type: change action type
        :param str credit_comment: comment for credit

        :return: None
        :rtype: None
        """
        self.conds = conds
        self.deposit = deposit
        self.change_type = change_type
        self.deposit_comment = deposit_comment



"""Change deposit extended API method."""
from ibsng.handler.handler import Handler


class changeDepositExtended(Handler):
    """Change deposit extended method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.user_id, int)
        self.is_valid(self.credit, float)
        self.is_valid(self.change_type, str)
        self.is_valid_content(self.change_type, ("ADD|SET|MULTIPLY"))
        self.is_valid(self.credit_comment, str, False)

    def setup(self, user_id, deposit, change_type, deposit_comment=""):
        """Setup required parameters.

        :param str user_id: ibsng user id
        :param float deposit: new deposit value
        :param choice change_type: type of change (ADD, SET, MULTIPLY)
        :param str deposit_comment: omment for this deposit

        :return: None
        :rtype: None
        """
        self.user_id = user_id
        self.deposit = deposit
        self.change_type = change_type
        self.deposit_comment = deposit_comment



"""Create session for failed user by ip API method."""
from ibsng.handler.handler import Handler


class createSessionForFailedUserByIP(Handler):
    """Create session for failed user by ip method class."""

    pass



"""Admin change password API method."""
from ibsng.handler.handler import Handler


class changePassword(Handler):
    """Admin change password class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.admin_username, str)
        self.is_valid(self.admin_password, str)

    def setup(self, admin_username, new_password):
        """Setup required parameters.

        :param str admin_username: admin authentication username
        :param str new_password: admin authentication new password

        :return: None
        :rtype: None
        """
        self.admin_username = admin_username
        self.new_password = new_password



"""Bulk action to renew user API method."""
from ibsng.handler.handler import Handler


class bulkRenewUsers(Handler):
    """Bulk action to renew user method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.conds, dict)
        self.is_valid(self.renew_comment, str, False)

    def setup(self, conds, renew_comment=""):
        """Setup required parameters.

        :param dict conds: conditions
        :param str renew_comment: comment for this action

        :return: None
        :rtype: None
        """
        self.conds = conds
        self.renew_comment = renew_comment



"""Invoice paid API method."""
from ibsng.handler.handler import Handler


class invoicePaid(Handler):
    """Invoice paid method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.invoice_id_list, list)
        self.is_valid(self.paid_amount_list, list)
        self.is_valid(self.payment_comment_list, list)

    def setup(self, invoice_id_list, paid_amount_list, payment_comment_list):
        """Setup required parameters.

        :param list invoice_id_list: list of Invoice IDs
        :param list paid_amount_list: list of paid amounts, length of list must
                                      be the same as `invoice_id_list`
        :param list payment_comment_list: list of comment strings, length of
                                          list must be the same as
                                          `invoice_id_list`

        :return: None
        :rtype: None
        """
        self.invoice_id_list = invoice_id_list
        self.paid_amount_list = paid_amount_list
        self.payment_comment_list = payment_comment_list



"""Get extra charge profile names API method."""
from ibsng.handler.handler import Handler


class getExtraChargeProfileNames(Handler):
    """Get extra charge profile names method class."""

    pass



"""Get invoice by id API method."""
from ibsng.handler.handler import Handler


class getInvoiceByID(Handler):
    """Get invoice by id method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.invoice_id, int)

    def setup(self, invoice_id):
        """Setup required parameters.

        :param int invoice_id: ibsng invoice id

        :return: None
        :rtype: None
        """
        self.invoice_id = invoice_id



"""List groups with id API method."""
from ibsng.handler.handler import Handler


class listGroupsWithIDs(Handler):
    """List groups with id method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.active_only, bool)

    def setup(self, active_only=False):
        """Setup required parameters.

        :param bool active_only: pass true if you only want active groups

        :return: None
        :rtype: None
        """
        self.active_only = active_only



"""Main request handler library.

This library prepares and send http request.
"""
from requests.exceptions import ReadTimeout

from ibsng.exception.connection import (ServerConnectionError,
                                        ServerConnectionTimeout)
from ibsng.util import hash


class Request:
    """Request handler class."""

    def __init__(self, server_addr, session, timeout):
        """Request handler init.

        :param server_addr: connection string
        :type server_addr: str
        :param session: connection session ID
        :type session: str
        """
        self._server_addr = server_addr
        self._session = session
        self._timeout = timeout

    def send(self, method, **params):
        """Send request to the server.

        :raises ValueError: raise ValueError when server result is not JSON

        :param method: handler and method names
        :type method: str
        :param **params: send method parameters
        :type **params: dict

        :return: server result
        :rtype: dict
        """
        data = hash.serialize(dict(method=method, params=params))
        # Post request
        try:
            result = self._session.post(url=self._server_addr,
                                        data=data, timeout=self._timeout)
        except ReadTimeout:
            raise ServerConnectionTimeout()
        except Exception as e:
            raise ServerConnectionError(e)

        # Convert server response to object
        try:
            return hash.deserialize(result.text)
        except Exception:
            raise ValueError("Convert server response failed.")



"""Check user existence API method."""
from ibsng.handler.handler import Handler


class doesUserExists(Handler):
    """Check user existence method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.normal_username, str)

    def setup(self, normal_username):
        """Setup required parameters.

        :param str normal_username: ibsng normal username

        :return: None
        :rtype: None
        """
        self.normal_username = normal_username



"""API methods interface handler.

All method classes should get instance of Interface class.
New method should override __init__ method and if you want
to use Python keywords in header keys like 'from', try to add
an underscore after that like so 'from_'. Interface
class will replace just one '_' with blank.

Note 1: New method name should be in camel case style and first
word must be small. Also it's file name should be small and use
underscore instead of camel case like so:

  'myNewMethod' -> my_new_method.py

Note 2: try to use default structure and avoid to override other
methods of Interface class.

Sample:

class myNewMethod(Interface):
    def __init__(self, from_, to_):
        self.from_ = from_
        self.to_ = to_
"""
from abc import ABCMeta

from ibsng.handler.initiator import Initiator
from ibsng.handler.control import Control


class Handler(Initiator, Control, metaclass=ABCMeta):
    """Base handler class to handle API methods."""

    _method_headers = dict()

    def __init__(self, *args, **kwargs):
        """Make headers ready."""
        super(Handler, self).__init__(*args, **kwargs)
        self.update_headers()
        self.control()

    def update_headers(self):
        """Update required headers keys before each request.

        :return: None
        :rtype: None
        """
        self._method_headers.update({k[:-1] if k.endswith('_') else k: v
                                     for k, v in vars(self).items()})

    def get_headers(self):
        """Return current method headers.

        :return: return method headers
        :rtype: dict
        """
        return self._method_headers

    def outcome(self, result):
        """Convert result items to required types.

        .. note: Should be implemented based on requirements

        :param dict result: IBSng json result

        :return: None
        :rtype: None
        """
        pass

    def serialize(self):
        """Serialize all entries and make ready to send.

        :return: method headers
        :rtype: dict
        """
        return self._method_headers



"""Performa invoice paid API method."""
from ibsng.handler.handler import Handler


class proformaInvoicePaid(Handler):
    """Performa invoice paid method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.pi_id_list, list)
        self.is_valid_content(self.pi_id_list, self.ID_PATTERN)
        self.is_valid(self.paid_amount_list, list)
        # FIXME: need validate amount value (int, float or ...)
        self.is_valid(self.payment_comment_list, list)

    def setup(self, pi_id_list, paid_amount_list, payment_comment_list):
        """Setup required parameters.

        :param list pi_id_list: list of proforma invoice ids
        :param list paid_amount_list: list of paid amounts, length of list must
                                      be the same as `pi_id_list`
        :param list payment_comment_list: list of comment strings, length of
                                          list must be the same as `pi_id_list`

        :return: None
        :rtype: None
        """
        self.pi_id_list = pi_id_list
        self.paid_amount_list = paid_amount_list
        self.payment_comment_list = payment_comment_list



"""Echo request input API method."""
from ibsng.handler.handler import Handler


class echo(Handler):
    """Echo request input method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.echo, str)

    def setup(self, echo):
        """Setup required parameters.

        :param str echo: string arg

        :return: None
        :rtype: None
        """
        self.echo = echo



"""Delete permission value API method."""
from ibsng.handler.handler import Handler


class delPermissionValue(Handler):
    """Delete permission value method class."""

    def control(self):
        """Validate inputs after method setup.

        :return: None
        :rtype: None
        """
        self.is_valid(self.admin_username, str)
        self.is_valid(self.perm_name, str)
        self.is_valid(self.perm_value, str)

    def setup(self, admin_username, perm_name, perm_value):
        """Setup required parameters.

        :param str admin_username: admin_username
        :param str perm_name: permission name
        :param str perm_value: permission value

        :return: None
        :rtype: None
        """
        self.admin_username = admin_username
        self.perm_name = perm_name
        self.perm_value = perm_value



"""Package setup."""
from setuptools import setup, find_packages

setup(
    name="pyibsng",
    version="1.2.5",
    description="IBSng python driver",
    long_description=open("README.rst").read(),
    author="Parspooyesh Fanavar",
    author_email="info@parspooyesh.com",
    url="https://github.com/ParspooyeshFanavar/pyibsng",
    license="MIT",
    keywords=["ibsng", "driver", "library", "json-rpc", "rpc"],
    install_requires=['requests'],
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Terminals",
        "Intended Audience :: Developers",
    ],
)



"""Update charge rule attributes API method."""
from ibsng.handler.handler import Handler


class updateChargeRuleAttrs(Handler):
    """Update charge rule attributes method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.charge_rule_id, int)
        self.is_valid(self.charge_name, str)
        self.is_valid(self.update_attrs, dict, False)
        self.is_valid(self.delete_attrs, list, False)

    def setup(self, charge_rule_id, charge_name,
              update_attrs={}, delete_attrs=[]):
        """Setup required parameters.

        :param int charge_rule_id: charge rule id
        :param str charge_name: charge name
        :param dict update_attrs: updated attributes. structure:
                                  {attribute_name: attribute_value}
        :param list delete_attrs: deleted attributes

        :return: None
        :rtype: None
        """
        self.charge_rule_id = charge_rule_id
        self.charge_name = charge_name
        self.update_attrs = update_attrs
        self.delete_attrs = delete_attrs



"""Search user API method."""
from ibsng.handler.handler import Handler


class searchUser(Handler):
    """Search user method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.conds, dict)
        self.is_valid(self.from_, int)
        self.is_valid(self.to_, int)
        self.is_valid(self.order_by, str)
        self.is_valid(self.desc, bool)

    def setup(self, conds, from_, to_, order_by, desc):
        """Setup required parameters.

        :param dict conds: conditions
        :param int from: start index, for pagination
        :param int to: end index, for pagination
        :param choice order_by: order by field
        :param bool desc: descending order

        :return: None
        :rtype: None
        """
        self.conds = conds
        self.from_ = from_
        self.to_ = to_
        self.order_by = order_by
        self.desc = desc



"""IBSng 'authentication' exceptions handler."""
from .base import IBSngException


class InvalidAuthentication(IBSngException):
    """Raise when user authentication failed."""

    _err_code = 501

    def __init__(self):
        """Raise when user authentication failed."""
        message = "Authentication failed. Check username or password."
        super(InvalidAuthentication, self).__init__(message)


class InvalidSessionId(IBSngException):
    """Raise when user session id is invalid."""

    _err_code = 502

    def __init__(self):
        """Raise when user session id is invalid."""
        message = "User session id is invalid. Login again."
        super(InvalidSessionId, self).__init__(message)



"""Get usernames API method."""
from ibsng.handler.handler import Handler


class getUsernames(Handler):
    """Get usernames method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.starts_with, str)

    def setup(self, starts_with):
        """Setup required parameters.

        :param str starts_with: starts with (name)

        :return: None
        :rtype: None
        """
        self.starts_with = starts_with



"""Get list of permission templates API method."""
from ibsng.handler.handler import Handler


class getListOfPermTemplates(Handler):
    """Get list of permission templates method class."""

    pass



"""Add new node API method."""
from ibsng.handler.handler import Handler


class addNode(Handler):
    """Add new node class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.hostname, str)
        self.is_valid(self.interface_name, str)
        self.is_valid(self.parent_id, int)
        self.is_valid_content(self.parent_id, self.ID_PATTERN)
        self.is_valid(self.rate_kbits, int)
        self.is_valid(self.ceil_kbits, int)
        self.is_valid(self.priority, int)
        self.is_valid_content(self.priority, self.POSITIVE_NUMBER)

    def setup(self, hostname, interface_name, parent_id,
              rate_kbits, ceil_kbits, priority):
        """Setup required parameters.

        :param str hostname: host name
        :param str interface_name: interface name
        :param int parent_id: parent node
        :param int rate_kbits: kilo bits
        :param int ceil_kbits: kilo bits
        :param int priority: node priority

        :return: None
        :rtype: None
        """
        self.hostname = hostname
        self.interface_name = interface_name
        self.parent_id = parent_id
        self.rate_kbits = rate_kbits
        self.ceil_kbits = ceil_kbits
        self.priority = priority



"""Delete saved report API method."""
from ibsng.handler.handler import Handler


class deleteSavedReport(Handler):
    """Delete saved report method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.report_auth, str)
        self.is_valid(self.report_type, str)
        self.is_valid(self.report_date, str)
        self.is_valid(self.report_status, str)
        self.is_valid_content(self.report_status, "(started|finished)")
        self.is_valid(self.report_format, str)
        self.is_valid_content(self.report_format, "(csv|pdf)")

    def setup(self, report_auth, report_type,
              report_date, report_status, report_format):
        """Setup required parameters.

        :param str report_auth: admin username
        :param str report_type: report type
        :param str report_date: format: YYYYMMDD
        :param str report_status: report status (started, finished)
        :param str report_format: report format (csv, pdf)

        :return: None
        :rtype: None
        """
        self.report_auth = report_auth
        self.report_type = report_type
        self.report_date = report_date
        self.report_status = report_status
        self.report_format = report_format



"""IBSng 'handler' exceptions handler."""
from .base import IBSngException


class HandlerNotFound(IBSngException):
    """Raise when handler directory not found."""

    _err_code = 301

    def __init__(self, handler_name):
        """Raise when handler directory not found.

        :param handler_name: handler name
        :type handler_name: str
        """
        message = "There is no '{}' handler in driver.".format(handler_name)
        super(HandlerNotFound, self).__init__(message)



"""Add IPto pool API method."""
from ibsng.handler.handler import Handler


class addIPtoPool(Handler):
    """Add IPto pool method class."""

    def control(self):
        """Validate inputs after method setup.

        :return: None
        :rtype: None
        """
        self.is_valid(self.ippool_name, str)
        self.is_valid(self.ip, str)

    def setup(self, ippool_name, ip):
        """Setup required parameters.

        :param str ippool_name: ip pool name
        :param str ip: ip address

        :return: None
        :rtype: None
        """
        self.ippool_name = ippool_name
        self.ip = ip



"""Load permissions from another admin API method."""
from ibsng.handler.handler import Handler


class loadPermsFromAnotherAdmin(Handler):
    """Load permissions from another admin method class."""

    def control(self):
        """Validate inputs after method setup.

        :return: None
        :rtype: None
        """
        self.is_valid(self.admin_username, str)
        self.is_valid(self.load_from_admin_username, str)

    def setup(self, admin_username, load_from_admin_username):
        """Setup required parameters.

        :param str admin_username: admin username
        :param str load_from_admin_username: load from admin username

        :return: None
        :rtype: None
        """
        self.admin_username = admin_username
        self.load_from_admin_username = load_from_admin_username



"""Get invoice by rule id API method."""
from ibsng.handler.handler import Handler


class getInvoiceRuleByID(Handler):
    """Get invoice by rule id method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.rule_id, int)

    def setup(self, rule_id):
        """Setup required parameters.

        :param int rule_id: ibsng rule id

        :return: None
        :rtype: None
        """
        self.rule_id = rule_id



"""Set ISP page style API method."""
from ibsng.handler.handler import Handler


class setISPPageStyle(Handler):
    """Set ISP page style method class."""

    def control(self):
        """Validate inputs after method setup.

        :return: None
        :rtype: None
        """
        self.is_valid(self.isp_id, int)
        self.is_valid(self.header_logo_ext, str)
        self.is_valid(self.header_logo_contents, str)
        self.is_valid(self.header_bg_color, str)
        self.is_valid(self.header_text_color, str)
        self.is_valid(self.footer_logo_ext, str)
        self.is_valid(self.footer_logo_contents, str)
        self.is_valid(self.footer_bg_color, str)
        self.is_valid(self.footer_text_color, str)
        self.is_valid(self.footer_msg, str)

    def setup(self, isp_id, header_logo_ext, header_logo_contents,
              header_bg_color, header_text_color, footer_logo_ext,
              footer_logo_contents, footer_bg_color, footer_text_color,
              footer_msg):
        """Setup required parameters.

        :param int isp_id: isp id
        :param str header_logo_ext: header logo file extension,
                                    for example 'gif'
        :param str header_logo_contents: base64-encoded data of header
                                         logo image
        :param str header_bg_color: header background color, in html color
                                    format, for example 'E0DEDE'
        :param str header_text_color: header text color, html color format,
                                      for example 'E0DEDE'
        :param str footer_logo_ext: footer logo file extension,
                                    for example 'gif'
        :param str footer_logo_contents: base64-encoded data of footer
                                         logo image
        :param str footer_bg_color: footer background color, in html color
                                    format, for example 'E0DEDE'
        :param str footer_text_color: footer text color, in html color format,
                                      for example 'E0DEDE'
        :param str footer_msg: base64-encoded string of footer message

        :return: None
        :rtype: None
        """
        self.isp_id = isp_id
        self.header_logo_ext = header_logo_ext
        self.header_logo_contents = header_logo_contents
        self.header_bg_color = header_bg_color
        self.header_text_color = header_text_color
        self.footer_logo_ext = footer_logo_ext
        self.footer_logo_contents = footer_logo_contents
        self.footer_bg_color = footer_bg_color
        self.footer_text_color = footer_text_color
        self.footer_msg = footer_msg



"""Delete LDAP IBSng mapping API method."""
from ibsng.handler.handler import Handler


class deleteLDAPIBSMapping(Handler):
    """Delete LDAP IBSng mapping method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.domain, str)
        self.is_valid(self.ldap_attribute, str)
        self.is_valid(self.IBS_field, str)
        self.is_valid(self.sync, bool)

    def setup(self, domain, ldap_attribute, ibs_field, sync):
        """Setup required parameters.

        :param str domain: name of ldap domain in which mapping takes place
        :param str ldap_attribute: ldap side of attribute mapping
        :param str ibs_field: ibs side of attribute mapping
        :param bool sync: specifies if the mapping be synchronized or not

        :return: None
        :rtype: None
        """
        self.domain = domain
        self.ldap_attribute = ldap_attribute
        self.IBS_field = ibs_field
        self.sync = sync



"""Bulk action to kill user API method."""
from ibsng.handler.handler import Handler


class bulkKillUsers(Handler):
    """Bulk action to kill user method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.conds, dict)
        self.is_valid(self.kill, bool)

    def setup(self, conds, kill=False):
        """Setup required parameters.

        :param dict conds: conditions
        :param bool kill: True means 'Kill User', False means 'Clear User'

        :return: None
        :rtype: None
        """
        self.conds = conds
        self.kill = kill



"""Issue performa invoice API method."""
from ibsng.handler.handler import Handler


class issueProformaInvoice(Handler):
    """Issue performa invoice method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.attrs_list, list)
        self.is_valid(self.arbitrary_amount, float)

    def setup(self, attrs_list, arbitrary_amount):
        """Setup required parameters.

        :param list attrs_list: list of attributes
        :param float arbitrary_amount: arbitrary amount

        :return: None
        :rtype: None
        """
        self.arbitrary_amount = arbitrary_amount
        self.attrs_list = attrs_list



"""Update node API method."""
from ibsng.handler.handler import Handler


class updateNode(Handler):
    """Update node class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.node_id, int)
        self.is_valid_content(self.node_id, self.ID_PATTERN)
        self.is_valid(self.rate_kbits, int)
        self.is_valid(self.ceil_kbits, int)
        self.is_valid(self.priority, int)
        self.is_valid_content(self.priority, self.POSITIVE_NUMBER)

    def setup(self, node_id, rate_kbits, ceil_kbits, priority):
        """Setup required parameters.

        :param int node_id: node ID
        :param int rate_kbits: new kilo bits
        :param int ceil_kbits: new kilo bits
        :param int priority: new priority

        :return: None
        :rtype: None
        """
        self.node_id = node_id
        self.rate_kbits = rate_kbits
        self.ceil_kbits = ceil_kbits
        self.priority = priority



"""Change notification status API method."""
from ibsng.handler.handler import Handler


class changeNotificationStatus(Handler):
    """Change notification status method class."""

    def control(self):
        """Validate inputs after setup method.

        :return: None
        :rtype: None
        """
        self.is_valid(self.notification_id, int)
        self.is_valid_content(self.notification_id, self.ID_PATTERN)
        self.is_valid(self.read, bool)

    def setup(self, notification_id, read):
        """Setup required parameters.

        :param int notification_id: ibsng notification id
        :param bool read: change notification 'read' status

        :return: None
        :rtype: None
        """
        self.notification_id = notification_id
        self.read = read
